# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

import numpy as np
import torch

from nncf.common.factory import TModel
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.transformations.commands import TransformationPriority
from nncf.common.graph.transformations.layout import TransformationLayout
from nncf.common.tensor_statistics.aggregator import StatisticPointsContainer
from nncf.common.tensor_statistics.aggregator import StatisticsAggregator
from nncf.torch.graph.transformations.commands import PTInsertionCommand
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.tensor import PTNNCFTensor
from nncf.torch.tensor_statistics.algo import create_register_input_hook


class PTStatisticsAggregator(StatisticsAggregator):
    HOOKS_GROUP_NAME = "statistics_hooks"

    def collect_statistics(self, model: NNCFNetwork, graph: NNCFGraph) -> None:
        with torch.no_grad():
            super().collect_statistics(model, graph)
        model.nncf.remove_hooks_group(self.HOOKS_GROUP_NAME)

    def _register_statistics(
        self, outputs: Dict[str, PTNNCFTensor], statistic_points: StatisticPointsContainer
    ) -> None:
        return

    def _get_transformation_layout_extra_outputs(
        self, statistic_points: StatisticPointsContainer
    ) -> TransformationLayout:
        transformation_layout = TransformationLayout()
        transformation_commands = []

        for _statistic_points in statistic_points.values():
            for _statistic_point in _statistic_points:
                for collectors in _statistic_point.algorithm_to_tensor_collectors.values():
                    for collector in collectors:
                        transformation_commands.append(
                            PTInsertionCommand(
                                _statistic_point.target_point,
                                create_register_input_hook(collector=collector),
                                TransformationPriority.FP32_TENSOR_STATISTICS_OBSERVATION,
                                hooks_group_name=self.HOOKS_GROUP_NAME,
                            )
                        )

        for transformation_command in transformation_commands:
            transformation_layout.register(transformation_command)

        return transformation_layout

    @staticmethod
    def _get_merged_statistic_points(
        statistic_points: StatisticPointsContainer, model: TModel, graph: NNCFGraph
    ) -> StatisticPointsContainer:
        # TODO: mirgate to experimental statistic collector and use common merging algorithm
        return statistic_points

    @staticmethod
    def _process_outputs(outputs: Dict[str, np.ndarray]) -> Dict[str, PTNNCFTensor]:
        return outputs



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

import tensorflow as tf


class CheckpointManagerCallback(tf.keras.callbacks.Callback):
    """Callback to save the NNCF TF compression model with compression state."""

    def __init__(self, checkpoint: tf.train.Checkpoint, directory: str, save_freq: Union[str, int] = "epoch"):
        """
        :param checkpoint: The `tf.train.Checkpoint` instance to save and manage
            checkpoints for.
        :param directory: The path to a directory in which to write checkpoints.
        :param save_freq: `'epoch'` or integer. When using `'epoch'`, the callback saves.
            the model after each epoch and name number of the checkpoint correspond to epoch
            model was saved. When using integer, the callback saves the model at end of this many
            batches and name number of the checkpoint correspond to number of passed batches * `save_freq`.
        """
        super().__init__()
        self._last_batch_seen = 0
        self._batches_seen_since_last_saving = 0
        if save_freq != "epoch" and not isinstance(save_freq, int):
            raise ValueError("Unrecognized save_freq: {}".format(save_freq))

        self._checkpoint_manager = tf.train.CheckpointManager(checkpoint, directory, None)
        self._save_freq = save_freq

    def on_train_batch_end(self, batch, logs=None):
        if self._should_save_on_batch(batch):
            self._save_model()

    def on_epoch_end(self, epoch, logs=None):
        if self._save_freq == "epoch":
            self._save_model()

    def _should_save_on_batch(self, batch):
        """Handles batch-level saving logic, supports steps_per_execution."""
        if self._save_freq == "epoch":
            return False

        if batch <= self._last_batch_seen:  # New epoch.
            add_batches = batch + 1  # batches are zero-indexed.
        else:
            add_batches = batch - self._last_batch_seen

        self._batches_seen_since_last_saving += add_batches
        self._last_batch_seen = batch

        if self._batches_seen_since_last_saving >= self._save_freq:
            self._batches_seen_since_last_saving = 0
            return True
        return False

    def _save_model(self):
        self._checkpoint_manager.save()



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import shutil
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List

import networkx as nx
import numpy as np
import torch

import nncf
from nncf.common.insertion_point_graph import InsertionPointGraph
from nncf.common.insertion_point_graph import InsertionPointGraphNodeType
from nncf.common.logging import nncf_logger
from nncf.common.utils.dot_file_rw import write_dot_graph
from nncf.common.utils.os import safe_open
from nncf.torch.debug import CallCountTracker
from nncf.torch.debug import DebugInterface
from nncf.torch.quantization.layers import QUANTIZATION_MODULES

if TYPE_CHECKING:
    from nncf.torch.nncf_network import NNCFNetwork


class QuantizationDebugInterface(DebugInterface):
    QUANTIZERS_IN_NNCF_MODULES_TRACKER_NAME = "quantized_modules"
    ACTIVATION_QUANTIZERS_TRACKER_NAME = "activation_quantizers"

    def __init__(self):
        self.call_trackers = {
            self.QUANTIZERS_IN_NNCF_MODULES_TRACKER_NAME: CallCountTracker(
                QuantizationDebugInterface.QUANTIZERS_IN_NNCF_MODULES_TRACKER_NAME
            ),
            self.ACTIVATION_QUANTIZERS_TRACKER_NAME: CallCountTracker(
                QuantizationDebugInterface.ACTIVATION_QUANTIZERS_TRACKER_NAME
            ),
        }
        self.graph_size = 0

        from nncf.common.utils.debug import DEBUG_LOG_DIR

        self.dump_dir = Path(DEBUG_LOG_DIR) / Path("debug_dumps")
        self.dump_dir.mkdir(parents=True, exist_ok=True)
        self.scale_dump_dir = self.dump_dir / Path("scale")
        if self.scale_dump_dir.exists():
            shutil.rmtree(str(self.scale_dump_dir))
        self.scale_dump_dir.mkdir(parents=True, exist_ok=True)
        self.forward_call_count = 0
        self._strict_forward = False

    def init_actual(self, owner_model: NNCFNetwork):
        from nncf.torch.graph.transformations.commands import ExtraCompressionModuleType

        quantization_types = [class_type.__name__ for class_type in QUANTIZATION_MODULES.registry_dict.values()]
        quantizers_in_nncf_modules = owner_model.nncf.get_modules_in_nncf_modules_by_type(quantization_types)
        nncf_module_quantizations_id_list: List[str] = [str(scope) for scope in quantizers_in_nncf_modules]

        activation_quantizer_id_list: List[str] = owner_model.nncf.get_compression_modules_by_type(
            ExtraCompressionModuleType.EXTERNAL_QUANTIZER
        ).keys()
        self.call_trackers[self.QUANTIZERS_IN_NNCF_MODULES_TRACKER_NAME].init_with_key_list(
            nncf_module_quantizations_id_list
        )
        self.call_trackers[self.ACTIVATION_QUANTIZERS_TRACKER_NAME].init_with_key_list(activation_quantizer_id_list)
        self._strict_forward = True

    def pre_forward_actions(self, module: NNCFNetwork):
        self.reset_counters()

    def post_forward_actions(self, module: NNCFNetwork):
        self.register_forward_call()

        ctx = module.nncf.get_tracing_context()
        self.set_graph_size(ctx.graph.get_nodes_count())

        quantization_types = [class_type.__name__ for class_type in QUANTIZATION_MODULES.registry_dict.values()]
        nncf_module_quantizations = module.nncf.get_modules_in_nncf_modules_by_type(quantization_types)

        for qm_scope, qm_module in nncf_module_quantizations.items():
            # Important - this will not work for DataParallel since it copies the
            # entire parent module for each thread and the `call_count` attributes
            # are incremented for thread local copies of `qm_module`, which are not
            # the same as the primary copies of `qm_module` iterated over at this point
            self.register_quantizer_module_call(str(qm_scope), qm_module.call_count)
            self.dump_scale(qm_module.get_trainable_params(), str(qm_scope))
            qm_module.reset_call_counter()
        self.print_call_stats()

        call_dict = ctx.get_node_call_counter_dict()
        total_calls = sum(call_dict.values())
        nncf_logger.debug(f"{total_calls} nodes called out of total {ctx.graph.get_nodes_count()}")
        if self._strict_forward:
            for tracker in self.call_trackers.values():
                if tracker.get_never_called_keys():
                    # This will always trigger for DataParallel - disregard or disable debug mode
                    # for DataParallel runs
                    raise nncf.InternalError(
                        f"{tracker.name} has never called modules: {tracker.get_never_called_keys()}!"
                    )

    def dump_scale(self, quantizer_scale_params: Dict[str, torch.Tensor], quantizer_name: str):
        import re

        quantizer_normalized_name = re.sub(r"[^\w\-_\. ]", "_", quantizer_name)
        for scale_param_name, scale_param in quantizer_scale_params.items():
            fname = "{}_{}.txt".format(quantizer_normalized_name, scale_param_name)
            with safe_open(self.scale_dump_dir / fname, "ab") as file:
                np.savetxt(file, scale_param.cpu().numpy().flatten())

    def reset_counters(self):
        for tracker in self.call_trackers.values():
            tracker.reset()

    def register_quantizer_module_call(self, key, counts=None):
        self.call_trackers[self.QUANTIZERS_IN_NNCF_MODULES_TRACKER_NAME].register_call(key, counts)

    def register_activation_quantize_call(self, key: str):
        self.call_trackers[self.ACTIVATION_QUANTIZERS_TRACKER_NAME].register_call(key)

    def print_call_stats(self):
        nncf_logger.debug(f" Graph size: {self.graph_size} nodes")
        for tracker in self.call_trackers.values():
            msg = f" {tracker.name} tracker:"
            msg += f" {tracker.get_total_call_count()} total calls;"

            never_called = tracker.get_never_called_keys()
            if never_called:
                msg += f" {len(never_called)} entries never called;"

            overcalled = tracker.get_overcalled_keys_with_call_counts()
            if overcalled:
                msg += f" {len(overcalled)} entries called more than once;"
            nncf_logger.debug(msg)

    def set_graph_size(self, new_size):
        if new_size != self.graph_size:
            nncf_logger.debug("\n")
            nncf_logger.debug(
                f"Warning - graph size has changed from {self.graph_size} to {new_size} since last forward"
            )
        self.graph_size = new_size

    def register_forward_call(self):
        self.forward_call_count += 1

    def visualize_insertion_point_graph(self, insertion_point_graph: InsertionPointGraph):
        out_graph = nx.MultiDiGraph()
        for node_key, node in insertion_point_graph.nodes.items():
            if node[InsertionPointGraph.NODE_TYPE_NODE_ATTR] in [
                InsertionPointGraphNodeType.PRE_HOOK,
                InsertionPointGraphNodeType.POST_HOOK,
            ]:
                target_point_data = node[InsertionPointGraph.INSERTION_POINT_NODE_ATTR]
                label = "TP: {}".format(str(target_point_data))
                out_graph.add_node(node_key, label=label, color="red")
            elif node[InsertionPointGraph.NODE_TYPE_NODE_ATTR] == InsertionPointGraphNodeType.OPERATOR:
                out_graph.add_node(node_key)
            else:
                raise nncf.InternalError("Invalid InsertionPointGraph node!")
        for u, v in insertion_point_graph.edges:
            out_graph.add_edge(u, v)

        for node_key, node in insertion_point_graph.nodes.items():
            if node[InsertionPointGraph.NODE_TYPE_NODE_ATTR] == InsertionPointGraphNodeType.OPERATOR:
                for ip_node_key in node[InsertionPointGraph.ASSOCIATED_IP_NODE_KEYS_NODE_ATTR]:
                    out_graph.add_edge(node_key, ip_node_key, style="dashed", headport="e", tailport="e")

        write_dot_graph(out_graph, self.dump_dir / Path("insertion_point_graph.dot"))



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from enum import IntEnum
from typing import Any, Dict, List, Optional, Tuple, TypeVar

from nncf.api.statistics import Statistics
from nncf.common.graph.transformations.layout import TransformationLayout
from nncf.common.utils.api_marker import api
from nncf.common.utils.backend import copy_model

TModel = TypeVar("TModel")


@api()
class CompressionLoss(ABC):
    """
    Used to calculate the additional loss to be added to the base loss during the
    training process. It uses the model graph to measure variables and activations
    values of the layers during the loss construction. For example, the $L_0$-based
    sparsity algorithm calculates the number of non-zero weights in convolutional
    and fully-connected layers to construct the loss function.
    """

    @abstractmethod
    def calculate(self, *args, **kwargs) -> Any:
        """
        Calculates and returns the compression loss value.
        """

    @abstractmethod
    def load_state(self, state: Dict[str, Any]) -> None:
        """
        Loads the compression loss state.

        :param state: The state of the compression loss, most likely obtained as the result of a `.get_state()` method
            call.
        """

    @abstractmethod
    def get_state(self) -> Dict[str, Any]:
        """
        Returns the compression loss state.
        """

    def __call__(self, *args, **kwargs) -> Any:
        """
        Calculates and returns the compression loss value. Same as `.calculate()`.
        """
        return self.calculate(*args, **kwargs)


@api()
class CompressionScheduler(ABC):
    """
    Implements the logic of compression method control during the training process.
    May change the method hyperparameters in regard to the current training step
    or epoch. For example, the sparsity method can smoothly increase the sparsity
    rate over several epochs.

    The `step()` and `epoch_step()` methods of the compression scheduler must be
    called at the beginning of each training step and epoch, respectively:

    ..  code-block:: python

        for epoch in range(0, num_epochs):
            scheduler.epoch_step()
            for i, (x, y) in enumerate(dataset):
                 scheduler.step()
                 ...

    """

    @abstractmethod
    def step(self, next_step: Optional[int] = None) -> None:
        """
        Should be called at the beginning of each training step to prepare
        the compression method to continue training the model in the `next_step`.

        :param next_step: The global step index for which the compression scheduler
            will update the state of the compression method.
        """

    @abstractmethod
    def epoch_step(self, next_epoch: Optional[int] = None) -> None:
        """
        Should be called at the beginning of each training epoch to prepare
        the compression method to continue training the model in the `next_epoch`.

        :param next_epoch: The epoch index for which the compression scheduler
            will update the state of the compression method.
        """

    @abstractmethod
    def load_state(self, state: Dict[str, Any]) -> None:
        """
        Loads the compression scheduler state, but does not update the state of the
        compression method.

        :param state: Output of `get_state()` method.
        """

    @abstractmethod
    def get_state(self) -> Dict[str, Any]:
        """
        Returns the compression scheduler state.
        """


@api()
class CompressionStage(IntEnum):
    """
    Specifies the compression stage for the model.
    """

    UNCOMPRESSED = 0
    PARTIALLY_COMPRESSED = 1
    FULLY_COMPRESSED = 2

    def __add__(self, other: "CompressionStage") -> "CompressionStage":
        """
        Defines compression stage of a composite compression controller, consist of
        two algorithms, where `self` is the compression stage of the first algorithm
        and other - compression stage of the second one.

        * ``UNCOMPRESSED         & UNCOMPRESSED         == UNCOMPRESSED``
        * ``PARTIALLY_COMPRESSED & PARTIALLY_COMPRESSED == PARTIALLY_COMPRESSED``
        * ``FULLY_COMPRESSED     & FULLY_COMPRESSED     == FULLY_COMPRESSED``
        * ``UNCOMPRESSED         & PARTIALLY_COMPRESSED == PARTIALLY_COMPRESSED``
        * ``UNCOMPRESSED         & FULLY_COMPRESSED     == PARTIALLY_COMPRESSED``
        * ``PARTIALLY_COMPRESSED & FULLY_COMPRESSED     == PARTIALLY_COMPRESSED``

        :param other: An instance of another compression stage.
        :return: The common compression stage of the two algorithms.
        """
        if self == other:
            return self
        return CompressionStage.PARTIALLY_COMPRESSED


@api()
class CompressionAlgorithmController(ABC):
    """
    A handle to the compression-specific modifications made to the model.
    Hosts entities that are to be used during the training process, such as compression scheduler and compression loss.

    :param target_model: The model with additional modifications necessary
        to enable algorithm-specific compression during fine-tuning built by the `CompressionAlgorithmBuilder`.
    """

    def __init__(self, target_model: TModel):
        self._model = target_model

    @property
    def model(self) -> TModel:
        """
        The compressed model object with which this controller is associated.
        """
        return self._model

    @property
    @abstractmethod
    def loss(self) -> CompressionLoss:
        """
        The compression loss for this particular algorithm combination.
        """

    @property
    @abstractmethod
    def scheduler(self) -> CompressionScheduler:
        """
        The compression scheduler for this particular algorithm combination.
        """

    @property
    @abstractmethod
    def name(self) -> str:
        """
        Name of the compression algorithm that is being controlled.
        Should be unique to identify the controller and its state among other controllers and their states.
        """

    @abstractmethod
    def load_state(self, state: Dict[str, Dict[str, Any]]) -> None:
        """
        Loads the compression controller state from the map of algorithm name to the dictionary with state attributes.

        :param state: map of the algorithm name to the dictionary with the corresponding state attributes.
        """

    @abstractmethod
    def get_state(self) -> Dict[str, Dict[str, Any]]:
        """
        Returns the compression controller state, which is the map of the algorithm name to the dictionary with the
        corresponding state attributes.
        """

    @abstractmethod
    def get_compression_state(self) -> Dict[str, Any]:
        """
        Returns the compression state - builder and controller state.
        This state should be used to unambiguously resume compression via `compression_state` argument of
        `create_compressed_model` method.

        :return: Compression state of the model to  resume compression from it.
        """

    def compression_stage(self) -> CompressionStage:
        """
        Returns the compression stage. Should be used on saving best checkpoints
        to distinguish between uncompressed, partially compressed, and fully
        compressed models.

        :return: The compression stage of the target model.
        """

    @abstractmethod
    def statistics(self, quickly_collected_only: bool = False) -> Statistics:
        """
        Returns a `Statistics` class instance that contains compression algorithm statistics.

        :param quickly_collected_only: Enables collection of the statistics that
            don't take too much time to compute. Can be helpful for the case when
            need to keep track of statistics on each training batch/step/iteration.
        """

    def strip_model(self, model: TModel, do_copy: bool = False) -> TModel:
        """
        Strips auxiliary layers that were used for the model compression, as it's
        only needed for training. The method is used before exporting the model
        in the target format.

        :param model: The compressed model.
        :param do_copy: Modify copy of the model, defaults to False.
        :return: The stripped model.
        """
        if do_copy:
            model = copy_model(model)
        return model

    def prepare_for_export(self) -> None:
        """
        Prepare the compressed model for exporting to a backend-specific model serialization format.
        """
        self._model = self.strip_model(self._model)

    def strip(self, do_copy: bool = True) -> TModel:
        """
        Returns the model object with as much custom NNCF additions as possible removed
        while still preserving the functioning of the model object as a compressed model.

        :param do_copy: If True (default), will return a copy of the currently associated model object. If False,
          will return the currently associated model object "stripped" in-place.
        :return: The stripped model.
        """
        return self.strip_model(self.model, do_copy)

    @abstractmethod
    def export_model(
        self,
        save_path: str,
        save_format: Optional[str] = None,
        input_names: Optional[List[str]] = None,
        output_names: Optional[List[str]] = None,
        model_args: Optional[Tuple[Any, ...]] = None,
    ) -> None:
        """
        Exports the compressed model to the specified format for deployment.

        Makes method-specific preparations of the model, (e.g. removing auxiliary
        layers that were used for the model compression), then exports the model to
        the specified path.

        :param save_path: The path where the model will be saved.
        :param save_format: Saving format. The default format will
            be used if `save_format` is not specified.
        :param input_names: Names to be assigned to the input tensors of the model.
        :param output_names: Names to be assigned to the output tensors of the model.
        :param model_args: Tuple of additional positional and keyword arguments
            which are required for the model's forward during export. Should be
            specified in the following format:

            * (a, b, {'x': None, 'y': y}) for positional and keyword arguments.
            * (a, b, {}) for positional arguments only.
            * ({'x': None, 'y': y},) for keyword arguments only.
        """

    @property
    @abstractmethod
    def compression_rate(self) -> float:
        """
        Returns a float compression rate value ranging from 0 to 1 (e.g. the sparsity level,
        or the ratio of filters pruned).
        """

    @compression_rate.setter
    @abstractmethod
    def compression_rate(self, compression_rate: float) -> None:
        """
        Set a float compression rate value in the model (e.g. the sparsity
        level or the ratio of filters pruned).

        :param compression_rate: The compressed rate value to be set.
        """

    @abstractmethod
    def disable_scheduler(self) -> None:
        """
        Disables current compression scheduler during training by changing it to a dummy one that does not change
        the compression rate.
        """

    @property
    @abstractmethod
    def maximal_compression_rate(self) -> float:
        """
        Returns the maximal model compression rate supported by the compression controller.
        """


@api()
class CompressionAlgorithmBuilder(ABC):
    """
    Determines which modifications should be made to the original model in order to enable algorithm-specific
    compression during fine-tuning.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """
        :return: name of the compression algorithm that is being built. Should be unique to identify the builder
        and its state among other builders and their states.
        """

    @abstractmethod
    def apply_to(self, model: TModel) -> TModel:
        """
        Applies algorithm-specific modifications to the model.

        :param model: The original uncompressed model.
        :return: The model with additional modifications necessary to enable
            algorithm-specific compression during fine-tuning.
        """

    @abstractmethod
    def build_controller(self, model: TModel) -> CompressionAlgorithmController:
        """
        Builds an instance of algorithm-specific `nncf.api.compression.CompressionAlgorithmController`
        to handle the additional modules, parameters, and hooks inserted into the model to enable algorithm-specific
        compression.

        :param model: The model with additional modifications necessary to enable
            algorithm-specific compression during fine-tuning.
        :return: The instance of a `CompressionAlgorithmController`-derived class, specific for this algorithm.
        """

    @abstractmethod
    def get_transformation_layout(self, model: TModel) -> TransformationLayout:
        """
        Computes necessary model transformations to enable algorithm-specific
        compression.

        :param model: The original uncompressed model.
        :return: The instance of the `TransformationLayout` class containing
            a list of algorithm-specific modifications.
        """

    @abstractmethod
    def initialize(self, model: TModel) -> None:
        """
        Initialize model parameters before training.

        :param model: The model with additional modifications necessary to enable
            algorithm-specific compression during fine-tuning.
        """

    @abstractmethod
    def load_state(self, state: Dict[str, Any]) -> None:
        """
        Initializes object from the supplied state.

        :param state: The state of the builder, most likely obtained as the result of a `.get_state()` call.
        """

    @abstractmethod
    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.
        """


class CompressionLevel(IntEnum):
    """
    Legacy class, now replaced by CompressionStage.
    Supports backward compatibility of older checkpoints produced with NNCF.
    CompressionLevel is deprecated and will be removed in future releases.
    """

    NONE = 0
    PARTIAL = 1
    FULL = 2

    @classmethod
    def map_legacy_level_to_stage(cls):
        return {
            CompressionLevel.NONE: CompressionStage.UNCOMPRESSED,
            CompressionLevel.PARTIAL: CompressionStage.PARTIALLY_COMPRESSED,
            CompressionLevel.FULL: CompressionStage.FULLY_COMPRESSED,
        }



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

import numpy as np
import openvino.runtime as ov

from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.openvino.engine import OVCompiledModelEngine
from nncf.openvino.graph.layer_attributes import OVLayerAttributes
from nncf.openvino.graph.metatypes.groups import CONSTANT_OPERATIONS
from nncf.openvino.graph.metatypes.groups import FAKE_QUANTIZE_OPERATIONS
from nncf.openvino.graph.metatypes.groups import INPUTS_QUANTIZABLE_OPERATIONS
from nncf.openvino.graph.metatypes.groups import OPERATIONS_WITH_WEIGHTS
from nncf.openvino.graph.metatypes.groups import QUANTIZE_AGNOSTIC_OPERATIONS
from nncf.openvino.graph.metatypes.groups import SHAPEOF_OPERATIONS
from nncf.openvino.graph.metatypes.openvino_metatypes import OVConcatMetatype
from nncf.openvino.graph.metatypes.openvino_metatypes import OVOpMetatype
from nncf.openvino.graph.model_utils import get_start_nodes_for_activation_path_tracing
from nncf.openvino.graph.model_utils import model_has_state
from nncf.openvino.graph.node_utils import get_bias_value
from nncf.openvino.graph.node_utils import get_weight_value
from nncf.openvino.graph.node_utils import is_node_with_bias
from nncf.quantization.algorithms.accuracy_control.backend import AccuracyControlAlgoBackend
from nncf.quantization.algorithms.accuracy_control.backend import PreparedModel


class OVPreparedModel(PreparedModel):
    """
    Implementation of the `PreparedModel` for OpenVINO backend.
    """

    def __init__(self, model: ov.Model):
        self._stateful = model_has_state(model)
        self._compiled_model = ov.compile_model(model)
        self._engine = None

    @property
    def model_for_inference(self) -> ov.CompiledModel:
        return self._compiled_model

    @property
    def engine(self) -> OVCompiledModelEngine:
        if self._engine is None:
            self._engine = OVCompiledModelEngine(self._compiled_model, self._stateful)
        return self._engine


class OVAccuracyControlAlgoBackend(AccuracyControlAlgoBackend):
    """
    Implementation of the `AccuracyControlAlgoBackend` for OpenVINO backend.
    """

    # Metatypes

    @staticmethod
    def get_op_with_weights_metatypes() -> List[OVOpMetatype]:
        return OPERATIONS_WITH_WEIGHTS

    @staticmethod
    def get_quantizer_metatypes() -> List[OVOpMetatype]:
        return FAKE_QUANTIZE_OPERATIONS

    @staticmethod
    def get_const_metatypes() -> List[OVOpMetatype]:
        return CONSTANT_OPERATIONS

    @staticmethod
    def get_quantizable_metatypes() -> List[OVOpMetatype]:
        return INPUTS_QUANTIZABLE_OPERATIONS

    @staticmethod
    def get_quantize_agnostic_metatypes() -> List[OVOpMetatype]:
        return QUANTIZE_AGNOSTIC_OPERATIONS + [OVConcatMetatype]

    @staticmethod
    def get_shapeof_metatypes() -> List[OVOpMetatype]:
        return SHAPEOF_OPERATIONS

    @staticmethod
    def get_start_nodes_for_activation_path_tracing(nncf_graph: NNCFGraph) -> List[NNCFNode]:
        return get_start_nodes_for_activation_path_tracing(nncf_graph)

    # Manipulations with bias value and weights

    @staticmethod
    def is_node_with_bias(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        return is_node_with_bias(node, nncf_graph)

    @staticmethod
    def is_node_with_weight(node: NNCFNode) -> bool:
        return node.metatype in OPERATIONS_WITH_WEIGHTS and isinstance(node.layer_attributes, OVLayerAttributes)

    @staticmethod
    def get_bias_value(node_with_bias: NNCFNode, nncf_graph: NNCFGraph, model: ov.Model) -> np.ndarray:
        return get_bias_value(node_with_bias, nncf_graph, model)

    @staticmethod
    def get_weight_value(node_with_weight: NNCFNode, model: ov.Model, port_id: int) -> np.ndarray:
        return get_weight_value(node_with_weight, model, port_id)

    @staticmethod
    def get_weight_tensor_port_ids(node: NNCFNode) -> List[Optional[int]]:
        return node.layer_attributes.get_const_port_ids()

    @staticmethod
    def get_model_size(model: ov.Model) -> int:
        model_size = 0
        for op in model.get_ops():
            if op.get_type_name() == "Constant":
                model_size += op.data.nbytes

        return model_size



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
from enum import IntEnum
from typing import Any, Dict, Optional

import torch

from nncf.common.logging import nncf_logger
from nncf.common.schedulers import BaseCompressionScheduler
from nncf.common.schedulers import PolynomialDecaySchedule
from nncf.config.schemata.experimental_schema import MOVEMENT_ENABLE_STRUCTURED_MASKING
from nncf.config.schemata.experimental_schema import MOVEMENT_FINAL_IMPORTANCE_THRESHOLD
from nncf.config.schemata.experimental_schema import MOVEMENT_POWER


class MovementSchedulerStage(IntEnum):
    """
    Describes the current stage of training with movement sparsity.
    """

    PRE_WARMUP = 0
    IN_WARMUP = 1
    POST_WARMUP = 2


class MovementSchedulerParams:
    """
    Stores the params to initialize the scheduler of movement sparsity.
    """

    def __init__(
        self,
        warmup_start_epoch: int,
        warmup_end_epoch: int,
        importance_regularization_factor: float,
        enable_structured_masking: bool = MOVEMENT_ENABLE_STRUCTURED_MASKING,
        init_importance_threshold: Optional[float] = None,
        final_importance_threshold: float = MOVEMENT_FINAL_IMPORTANCE_THRESHOLD,
        power: float = MOVEMENT_POWER,
        steps_per_epoch: Optional[int] = None,
    ):
        """
        Initializes and validates the params for scheduler.

        :param warmup_start_epoch: Index of the starting epoch (inclusive) for warmup stage.
        :param warmup_end_epoch: Index of the end epoch (exclusive) for warmup stage.
        :param importance_regularization_factor: The regularization factor on weight importance scores.
        :param enable_structured_masking: Whether to do structured mask resolution after warmup stage.
        :param init_importance_threshold: The initial value of importance threshold during warmup stage.
        :param final_importance_threshold: The final value of importance threshold during warmup stage.
        :param power: The power value of polynomial decay for threshold update during warmup stage.
        :param steps_per_epoch: Number of training steps in one epoch.
        """

        if steps_per_epoch is None and warmup_start_epoch < 1:
            raise ValueError(
                "`warmup_start_epoch` must be >= 1 to enable the auto calculation of "
                "`steps_per_epoch`. Please either change `warmup_start_epoch` to a larger "
                "number or specify `steps_per_epoch` in the config."
            )

        if warmup_start_epoch < 0 or warmup_end_epoch <= warmup_start_epoch:
            raise ValueError("Movement sparsity requires 0 <= warmup_start_epoch < warmup_end_epoch.")

        if importance_regularization_factor < 0:
            raise ValueError("`importance_regularization_factor` should not be a negative number.")

        if init_importance_threshold is not None and init_importance_threshold >= final_importance_threshold:
            nncf_logger.warning(
                "`init_importance_threshold` is equal to or greater than "
                "`final_importance_threshold`. Movement sparsity may not work as expected."
            )

        self.warmup_start_epoch = warmup_start_epoch
        self.warmup_end_epoch = warmup_end_epoch
        self.importance_regularization_factor = importance_regularization_factor
        self.enable_structured_masking = enable_structured_masking
        self.init_importance_threshold = init_importance_threshold
        self.final_importance_threshold = final_importance_threshold
        self.power = power
        self.steps_per_epoch = steps_per_epoch

    @classmethod
    def from_dict(cls, params: Dict[str, Any]) -> "MovementSchedulerParams":
        """
        Initialize `MovementSchedulerParams` object from the config in dict format.

        :param params: A dict that specifies the parameters of movement sparsity scheduler.
        :return: A `MovementSchedulerParams` object that stores the parameters from `params`.
        """
        warmup_start_epoch: int = params.get("warmup_start_epoch")
        warmup_end_epoch: int = params.get("warmup_end_epoch")
        importance_regularization_factor: float = params.get("importance_regularization_factor")
        enable_structured_masking: bool = params.get("enable_structured_masking", MOVEMENT_ENABLE_STRUCTURED_MASKING)
        init_importance_threshold: Optional[float] = params.get("init_importance_threshold")
        final_importance_threshold: float = params.get(
            "final_importance_threshold", MOVEMENT_FINAL_IMPORTANCE_THRESHOLD
        )
        power: float = params.get("power", MOVEMENT_POWER)
        steps_per_epoch: Optional[int] = params.get("steps_per_epoch")

        if None in [warmup_start_epoch, warmup_end_epoch, importance_regularization_factor]:
            raise ValueError(
                "`warmup_start_epoch`, `warmup_start_epoch` and `importance_regularization_factor` "
                "are required in config for Movement Sparsity."
            )

        return cls(
            warmup_start_epoch=warmup_start_epoch,
            warmup_end_epoch=warmup_end_epoch,
            importance_regularization_factor=importance_regularization_factor,
            enable_structured_masking=enable_structured_masking,
            init_importance_threshold=init_importance_threshold,
            final_importance_threshold=final_importance_threshold,
            power=power,
            steps_per_epoch=steps_per_epoch,
        )


class MovementPolynomialThresholdScheduler(BaseCompressionScheduler):
    """
    Movement Sparsity scheduler with a polynomial decay schedule for importance
    threshold and importance regularization factor.

    The scheduler will update importance threshold and importance regularization
    factor per optimizer step. Parameter `steps_per_epoch` should be provided
    in config for the per step calculation. If not provided, the scheduler will
    use the first epoch to calculate `steps_per_epoch` parameter. In this case,
    parameter `warmup_start_epoch` must be equal to or larger than 1, and the
    scheduler will start calculation only after `steps_per_epoch` is calculated.
    """

    def __init__(self, controller: "MovementSparsityController", params: MovementSchedulerParams):  # noqa: F821
        """
        Initializes a movement sparsity scheduler with a polynomial decay schedule.

        :param controller: Movement sparsity algorithm controller.
        :param params: Parameters for the scheduler.
        """
        super().__init__()
        self._controller = controller
        self._params = params
        self._schedule = PolynomialDecaySchedule(
            initial_value=0.0,
            target_value=1.0,
            target_epoch=(self._params.warmup_end_epoch - self._params.warmup_start_epoch),
            power=self._params.power,
            concave=True,
        )
        self._steps_per_epoch = self._params.steps_per_epoch
        self._init_importance_threshold = self._params.init_importance_threshold
        self._cached_importance_threshold_lambda = None
        self._is_controller_frozen = False
        self._steps_in_current_epoch = 0
        self._should_skip = False

    @property
    def current_stage(self) -> MovementSchedulerStage:
        if self._steps_per_epoch is None or self.current_step < int(
            self._params.warmup_start_epoch * self._steps_per_epoch
        ):
            return MovementSchedulerStage.PRE_WARMUP
        if self.current_step < int(self._params.warmup_end_epoch * self._steps_per_epoch):
            return MovementSchedulerStage.IN_WARMUP
        return MovementSchedulerStage.POST_WARMUP

    @property
    def current_importance_regularization_factor(self) -> float:
        """
        Calculates the value of importance regularization factor per the current state. The factor
        stays zero before warmup stage, and gradually increases during warmup, and stays at
        the fixed value after warmup.

        :return: The value of importance regularization factor at the current step.
        """
        current_stage = self.current_stage
        if current_stage == MovementSchedulerStage.PRE_WARMUP:
            return 0.0
        if current_stage == MovementSchedulerStage.IN_WARMUP:
            return self._calc_current_scheduled_value(0.0, self._params.importance_regularization_factor)
        return self._params.importance_regularization_factor

    @property
    def current_importance_threshold(self) -> float:
        """
        Calculates the value of importance threshold per the current state. The threshold
        stays `-math.inf` before warmup stage, and gradually increases from `self._init_importance_threshold`
        during warmup, and finally stays fixed at the specified final value.

        :return: The value of importance threshold at the current step.
        """
        current_stage = self.current_stage
        if current_stage == MovementSchedulerStage.PRE_WARMUP:
            return -math.inf
        if current_stage == MovementSchedulerStage.IN_WARMUP:
            assert self._init_importance_threshold is not None
            return self._calc_current_scheduled_value(
                self._init_importance_threshold, self._params.final_importance_threshold
            )
        return self._params.final_importance_threshold

    @property
    def enable_structured_masking(self) -> bool:
        return self._params.enable_structured_masking

    def epoch_step(self, next_epoch: Optional[int] = None) -> None:
        super().epoch_step(next_epoch)
        self._maybe_should_skip()
        self._steps_in_current_epoch = 0

    def step(self, next_step: Optional[int] = None) -> None:
        super().step(next_step)
        self._steps_in_current_epoch += 1
        if self._should_skip:
            return
        self._schedule_controller()

    def get_state(self) -> Dict[str, Any]:
        state = super().get_state()
        state["_init_importance_threshold"] = self._init_importance_threshold
        state["_steps_per_epoch"] = self._steps_per_epoch
        return state

    def load_state(self, state: Dict[str, Any]) -> None:
        super().load_state(state)
        self._init_importance_threshold = state["_init_importance_threshold"]
        self._steps_per_epoch = state["_steps_per_epoch"]
        if self._steps_per_epoch is None:  # It is the first epoch and `steps_per_epoch` not specified
            self._steps_in_current_epoch = self._current_step + 1
            self._should_skip = True
        else:
            self._steps_in_current_epoch = self._current_step % self._steps_per_epoch + 1

    def _schedule_controller(self):
        """
        Asks and updates the controller during training steps. It (1) updates the importance threshold
        and importance regularization factor in the operand at each step; (2) freezes the controller
        after warmup; (3) calculates the initial importance threshold if unspecified; (4) conducts
        structured masking if supported.
        """
        if self._init_importance_threshold is None and self.current_stage == MovementSchedulerStage.IN_WARMUP:
            adaptive_init_threshold = self._calc_init_threshold_from_controller(target_sparsity=0.001)
            nncf_logger.info(
                "Movement sparsity automatically calculates `init_importance_threshold` as "
                f"{adaptive_init_threshold} so that warmup starts from ~0.1% linear layer sparsity."
            )
            if adaptive_init_threshold >= self._params.final_importance_threshold:
                nncf_logger.warning(
                    "The auto-calculated `init_importance_threshold` is equal to or greater than "
                    "`final_importance_threshold`. Movement sparsity may not work as expected."
                )
            self._init_importance_threshold = adaptive_init_threshold
        if self.current_stage == MovementSchedulerStage.POST_WARMUP and (not self._is_controller_frozen):
            if self._params.enable_structured_masking:
                self._controller.reset_independent_structured_mask()
                self._controller.resolve_structured_mask()
                self._controller.populate_structured_mask()
            self._controller.freeze()
            self._is_controller_frozen = True
        self._update_operand_importance_threshold_and_factor()

    def _calc_current_scheduled_value(self, start_value: float, end_value: float) -> float:
        assert self.current_stage == MovementSchedulerStage.IN_WARMUP
        assert self._steps_per_epoch is not None
        schedule_current_step = self.current_step - int(self._params.warmup_start_epoch * self._steps_per_epoch)
        schedule_epoch = schedule_current_step // self._steps_per_epoch
        schedule_step = schedule_current_step % self._steps_per_epoch
        scale = self._schedule(schedule_epoch, schedule_step, self._steps_per_epoch)
        return start_value + scale * (end_value - start_value)

    @torch.no_grad()
    def _calc_init_threshold_from_controller(self, target_sparsity: float = 0.001) -> float:
        # Calculate the k-th smallest value over all importance scores as the initial importance threshold
        # so that roughly k weight elements are masked and thus target sparsity is achieved. We conduct this on
        # CPU to (1) limit GPU memory usage; (2) avoid the non-deterministic behavior of `torch.Tensor.kthvalue`
        # on CUDA.
        assert 0.0 <= target_sparsity < 1.0
        importance_tensors = []
        for minfo in self._controller.sparsified_module_info:
            operand = minfo.operand
            weight = operand.get_importance(is_bias=False, expanded=True)
            importance_tensors.append(weight.detach().cpu().view(-1))
            if operand.prune_bias:
                bias = operand.get_importance(is_bias=True, expanded=True)
                importance_tensors.append(bias.detach().cpu().view(-1))
        all_importances = torch.cat(importance_tensors)
        k = min(all_importances.numel(), int(all_importances.numel() * target_sparsity) + 1)  # k starts from 1
        return all_importances.kthvalue(k).values.item()

    def _update_operand_importance_threshold_and_factor(self):
        current = (self.current_importance_threshold, self.current_importance_regularization_factor)
        if current != self._cached_importance_threshold_lambda:
            for minfo in self._controller.sparsified_module_info:
                minfo.operand.importance_threshold = self.current_importance_threshold
                minfo.operand.importance_regularization_factor = self.current_importance_regularization_factor
        self._cached_importance_threshold_lambda = current

    def _maybe_should_skip(self) -> None:
        """
        Checks if the first epoch (with index 0) should be skipped to calculate
        the steps per epoch. If the skip is needed, then the internal state
        of the scheduler object will not be changed.
        """
        self._should_skip = False

        if self._steps_per_epoch is None and self._steps_in_current_epoch > 0:
            self._steps_per_epoch = self._steps_in_current_epoch

        if (
            self._steps_per_epoch is not None
            and self._steps_in_current_epoch > 0
            and self._steps_per_epoch != self._steps_in_current_epoch
        ):
            raise Exception(
                "Actual steps per epoch and steps per epoch from the scheduler "
                "parameters are different. Scheduling may be incorrect."
            )

        if self._steps_per_epoch is None:
            self._should_skip = True
            nncf_logger.info(
                "Movement sparsity scheduler updates importance threshold and regularization"
                "factor per optimizer step, but steps_per_epoch was not set in config. Will "
                "measure the actual steps per epoch as signaled by a .epoch_step() call."
            )



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from typing import Dict, List, Optional, Tuple, TypeVar

import nncf
from nncf import Dataset
from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.logging.track_progress import track
from nncf.common.tensor_statistics.statistic_point import StatisticPointsContainer
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import get_backend
from nncf.experimental.tensor import Tensor
from nncf.experimental.tensor import functions as fns
from nncf.experimental.tensor.definitions import TensorDataType
from nncf.parameters import CompressWeightsMode
from nncf.quantization.algorithms.layerwise.engine import LayerwiseEngine
from nncf.quantization.algorithms.weight_compression.backend import WeightCompressionAlgoBackend
from nncf.quantization.algorithms.weight_compression.config import WeightCompressionConfig
from nncf.quantization.algorithms.weight_compression.config import WeightCompressionParameters
from nncf.quantization.algorithms.weight_compression.weight_lowering import calculate_integer_quantization_params
from nncf.quantization.algorithms.weight_compression.weight_lowering import calculate_nf4_scale
from nncf.quantization.algorithms.weight_compression.weight_lowering import calculate_nf4_weight
from nncf.quantization.algorithms.weight_compression.weight_lowering import calculate_quantized_weight
from nncf.quantization.algorithms.weight_compression.weight_lowering import decompress_nf4_weight
from nncf.quantization.algorithms.weight_compression.weight_lowering import do_dequantization

TModel = TypeVar("TModel")


class GPTQ:
    """
    GPTQ algorithm implementation
    """

    def __init__(
        self,
        damp_percent: float = 0.1,
        block_size: int = 128,
        subset_size: int = 128,
    ):
        """
        :param damp_percent: The percent of the average Hessian diagonal to use for dampening,
            recommended value is 0.1.
        :param block_size: The size of the blocks used during quantization. Defaults to 128.
        :param subset_size: Number of data samples to calculate Hessian. Defaults to 128.
        """
        self._damp_percent = damp_percent
        self._block_size = block_size
        self._subset_size = subset_size
        self._backend = None
        self._backend_entity = None

        self._layerwise_engine = LayerwiseEngine(subset_size=self._subset_size)

    def _set_backend_entity(self, model: TModel) -> None:
        """
        Creates a helper class with a backend-specific logic of the algorithm.

        :param model: Backend-specific input model.
        """
        self._backend = get_backend(model)
        if self._backend == BackendType.OPENVINO:
            from nncf.quantization.algorithms.weight_compression.openvino_backend import OVWeightCompressionAlgoBackend

            self._backend_entity = OVWeightCompressionAlgoBackend(model)
        else:
            raise nncf.UnsupportedBackendError(
                f"Cannot return backend-specific entity because {self._backend.value} is not supported!"
            )

    def apply(
        self,
        model: TModel,
        graph: NNCFGraph,
        dataset: Dataset,
        weight_compression_parameters: List[WeightCompressionParameters],
        statistic_points: Optional[StatisticPointsContainer] = None,
        backend_entity: Optional[WeightCompressionAlgoBackend] = None,
    ) -> Tuple[TModel, Dict[str, Tensor], Dict[str, Tensor]]:
        """
        Applies the GPTQ algorithm to quantize the weights of the given model.

        :param model: The model to quantize.
        :param graph: The model graph.
        :param dataset: The dataset to use for quantization.
        :param weight_compression_parameters: Parameters for weight compression.
        :param statistic_points: Optional container for statistic points.
        :param backend_entity: Weight compression algorithm backend.
        :return: The quantized model and its scales and zero points.
        """
        self._backend_entity = backend_entity
        if self._backend_entity is None:
            self._set_backend_entity(model)

        scales = {}
        zero_points = {}

        target_nodes = []
        target_nodes_wc_params_map = {}
        matmul_metatypes = self._backend_entity.matmul_metatypes
        for wc_params in weight_compression_parameters:
            if wc_params.node_with_weight.metatype in matmul_metatypes:
                target_nodes.append(wc_params.node_with_weight)
                target_nodes_wc_params_map[wc_params.node_with_weight] = wc_params

        target_node_iterator = self._layerwise_engine.create_iterator_through_target_nodes(
            model, graph, target_nodes, dataset, statistic_points
        )
        for node, inputs in track(target_node_iterator, total=len(target_nodes), description="Applying GPTQ"):
            wc_params = target_nodes_wc_params_map[node]
            if wc_params.compression_config.group_size == -1:
                continue
            assert len(inputs) == 1
            _, input_tensors = next(iter(inputs.items()))
            hessian = self._calculate_hessian(node, input_tensors)
            scale, zero_point = self._quantize_weights(model, graph, wc_params, hessian)
            scales[wc_params.weight_name] = scale
            zero_points[wc_params.weight_name] = zero_point

        return model, scales, zero_points

    def get_statistic_points(
        self,
        model: TModel,
        graph: NNCFGraph,
        target_nodes: List[NNCFNode],
        backend_entity: Optional[WeightCompressionAlgoBackend] = None,
    ) -> StatisticPointsContainer:
        """
        Returns statistic points, for which StatisticsCollector should collect statistics.

        :param model: The model for statistics collection.
        :param graph: The model graph.
        :param backend_entity: Weight compression algorithm backend.
        :return: Statistic points, for which StatisticsCollector should collect statistics.
        """
        self._backend_entity = backend_entity
        if self._backend_entity is None:
            self._set_backend_entity(model)

        matmul_metatypes = self._backend_entity.matmul_metatypes
        filtered_nodes = []
        for node in target_nodes:
            if node.metatype in matmul_metatypes:
                filtered_nodes.append(node)

        return self._layerwise_engine.get_statistic_points(model, graph, filtered_nodes)

    def _calculate_hessian(self, node: NNCFNode, inputs: List[Tensor]) -> Tensor:
        """
        Calculates the Hessian matrix for the given node and inputs.

        :param node: The target node for Hessian calculation.
        :param inputs: List of input tensors.
        :return: The Hessian matrix as a tensor.
        """
        nsamples = 0

        if node.metatype in self._backend_entity.convolution_metatypes:
            raise RuntimeError("Convolution metatypes are not supported")
        if node.layer_attributes.input_attributes["transpose"]:
            raise RuntimeError("Transpose is not supported")

        hessian = fns.zeros(
            (inputs[0].shape[-1], inputs[0].shape[-1]), backend=inputs[0].backend, dtype=TensorDataType.float32
        )

        for inp in inputs:
            batch_size = 1 if len(inp.shape) == 2 else inp.shape[0]
            if node.metatype in self._backend_entity.matmul_metatypes:
                if len(inp.shape) == 3:
                    inp = inp.reshape((-1, inp.shape[-1]))
                inp = fns.transpose(inp)
            hessian *= nsamples / (nsamples + batch_size)
            nsamples += batch_size
            inp = fns.astype(inp, TensorDataType.float32) * math.sqrt(2 / nsamples)
            hessian += fns.matmul(inp, fns.transpose(inp))

        return hessian

    def _quantize_weights(
        self, model: TModel, graph: NNCFGraph, wc_params: WeightCompressionParameters, hessian: Tensor
    ):
        """
        Quantizes the weights of the model based on the calculated Hessian matrix.

        :param model: The model to quantize.
        :param graph: The model graph.
        :param wc_params: Parameters for weight compression.
        :param hessian: The Hessian matrix.
        :return: Scales and zero points used for quantization.
        """
        if wc_params.node_with_weight.metatype in self._backend_entity.convolution_metatypes:
            raise RuntimeError("Convolution metatypes are not supported")
        if not wc_params.node_with_weight.layer_attributes.constant_attributes[wc_params.weight_port_id]["transpose"]:
            raise RuntimeError("Transpose is not supported")

        weight_tensor = self._backend_entity.get_weight(
            wc_params.node_with_weight, wc_params.weight_port_id, model, graph
        )
        weight_tensor = fns.astype(weight_tensor, TensorDataType.float32)

        dead_indices = fns.diag(hessian) == 0
        hessian[dead_indices, dead_indices] = 1
        weight_tensor[:, dead_indices] = 0

        scales = []
        zero_points = []

        losses = fns.zeros_like(weight_tensor)
        quantized_tensor = fns.zeros_like(weight_tensor)

        columns = hessian.shape[0]
        group_size = wc_params.compression_config.group_size
        reduction_axes = wc_params.reduction_axes
        block_compression_config = WeightCompressionConfig(mode=wc_params.compression_config.mode)

        damp = self._damp_percent * fns.mean(fns.diag(hessian))
        diag_indices = fns.arange(columns, backend=hessian.backend, device=hessian.device)
        hessian[diag_indices, diag_indices] += damp
        hessian = fns.linalg.cholesky(hessian)
        hessian = fns.linalg.cholesky_inverse(hessian)
        hessian = fns.linalg.cholesky(hessian, upper=True)
        hessian_inv = hessian

        for i1 in range(0, columns, self._block_size):
            i2 = min(i1 + self._block_size, columns)
            count = i2 - i1

            weight_block = weight_tensor[:, i1:i2].clone()
            quantized_block = fns.zeros_like(weight_block)
            error_block = fns.zeros_like(weight_block)
            loss_block = fns.zeros_like(weight_block)
            hessian_inv_block = hessian_inv[i1:i2, i1:i2]

            for i in range(count):
                weight_col = weight_block[:, i]
                hessian_diag_val = hessian_inv_block[i, i]

                if group_size != -1 and (i1 + i) % group_size == 0:
                    if block_compression_config.mode == CompressWeightsMode.NF4:
                        scale = calculate_nf4_scale(weight_tensor[:, (i1 + i) : (i1 + i + group_size)], reduction_axes)
                        scales.append(scale)
                    else:
                        scale, zero_point = calculate_integer_quantization_params(
                            weight_tensor[:, (i1 + i) : (i1 + i + group_size)], reduction_axes, block_compression_config
                        )
                        scales.append(scale)
                        zero_points.append(zero_point)
                if block_compression_config.mode == CompressWeightsMode.NF4:
                    compressed_weights = calculate_nf4_weight(fns.unsqueeze(weight_col, 1), scales[-1])
                    quantized_col = decompress_nf4_weight(compressed_weights, scales[-1])
                else:
                    compressed_weights = calculate_quantized_weight(
                        fns.unsqueeze(weight_col, 1), scales[-1], zero_points[-1], block_compression_config
                    )
                    quantized_col = do_dequantization(compressed_weights, scales[-1], zero_points[-1])
                quantized_col = fns.flatten(quantized_col)
                quantized_block[:, i] = quantized_col
                loss_block[:, i] = (weight_col - quantized_col) ** 2 / hessian_diag_val**2

                error_col = (weight_col - quantized_col) / hessian_diag_val
                weight_block[:, i:] -= fns.matmul(
                    fns.unsqueeze(error_col, 1), fns.unsqueeze(hessian_inv_block[i, i:], 0)
                )
                error_block[:, i] = error_col

            quantized_tensor[:, i1:i2] = quantized_block
            losses[:, i1:i2] = loss_block / 2

            weight_tensor[:, i2:] -= fns.matmul(error_block, hessian_inv[i1:i2, i2:])

        quantized_tensor = quantized_tensor.reshape(weight_tensor.shape).astype(weight_tensor.dtype)
        self._backend_entity.set_weight(
            wc_params.node_with_weight, wc_params.weight_port_id, model, graph, quantized_tensor
        )

        scales = fns.stack(scales, axis=1)
        if wc_params.compression_config.mode == CompressWeightsMode.NF4:
            zero_points = None
        elif wc_params.compression_config.mode in [
            CompressWeightsMode.INT8_SYM,
            CompressWeightsMode.INT4_SYM,
        ]:
            zero_points = fns.squeeze(zero_points[0])
        else:
            zero_points = fns.stack(zero_points, axis=1)
        return scales, zero_points



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class api:
    API_MARKER_ATTR = "_nncf_api_marker"
    CANONICAL_ALIAS_ATTR = "_nncf_canonical_alias"

    def __init__(self, canonical_alias: str = None):
        self._canonical_alias = canonical_alias

    def __call__(self, obj):
        # The value of the marker will be useful in determining
        # whether we are handling a base class or a derived one.
        setattr(obj, api.API_MARKER_ATTR, obj.__name__)
        if self._canonical_alias is not None:
            setattr(obj, api.CANONICAL_ALIAS_ATTR, self._canonical_alias)
        return obj


def is_api(obj) -> bool:
    return hasattr(obj, api.API_MARKER_ATTR)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time

from torch import nn

from nncf.common.logging import nncf_logger
from nncf.config.schemata.defaults import PRUNING_LEGR_GENERATIONS
from nncf.config.schemata.defaults import PRUNING_LEGR_MAX_PRUNING
from nncf.config.schemata.defaults import PRUNING_LEGR_MUTATE_PERCENT
from nncf.config.schemata.defaults import PRUNING_LEGR_NUM_SAMPLES
from nncf.config.schemata.defaults import PRUNING_LEGR_POPULATION_SIZE
from nncf.config.schemata.defaults import PRUNING_LEGR_RANDOM_SEED
from nncf.config.schemata.defaults import PRUNING_LEGR_SIGMA_SCALE
from nncf.config.schemata.defaults import PRUNING_LEGR_TRAIN_STEPS
from nncf.torch.pruning.filter_pruning.global_ranking.evolutionary_optimization import EvolutionOptimizer
from nncf.torch.pruning.filter_pruning.global_ranking.evolutionary_optimization import LeGREvolutionEnv
from nncf.torch.pruning.filter_pruning.global_ranking.evolutionary_optimization import LeGRPruner
from nncf.torch.structures import LeGRInitArgs


class LeGR:
    """
    Class for training global ranking coefficients with Evolution optimization agent (but this agent can be easily
    replaced by any other RL agent with a similar interface) and LeGR-optimization environment.
    """

    def __init__(
        self,
        pruning_ctrl: "FilterPruningController",  # noqa: F821
        target_model: nn.Module,
        legr_init_args: LeGRInitArgs,
        train_steps: int = PRUNING_LEGR_TRAIN_STEPS,
        generations: int = PRUNING_LEGR_GENERATIONS,
        max_pruning: float = PRUNING_LEGR_MAX_PRUNING,
        random_seed: int = PRUNING_LEGR_RANDOM_SEED,
        population_size: int = PRUNING_LEGR_POPULATION_SIZE,
        num_samples: int = PRUNING_LEGR_NUM_SAMPLES,
        mutate_percent: float = PRUNING_LEGR_MUTATE_PERCENT,
        scale_sigma: float = PRUNING_LEGR_SIGMA_SCALE,
    ):
        """
        Initializing all necessary structures for optimization- LeGREvolutionEnv environment and EvolutionOptimizer
         agent.
        :param pruning_ctrl: pruning controller, an instance of FilterPruningController class
        :param target_model: model for which layers ranking coefficient will be trained
        :param legr_init_args: initial arguments for LeGR algorithm
        :param train_steps: number of training steps to evaluate accuracy of some ranking coefficients (action of agent)
        :param generations: number of generations in evolution algorithm optimization
        :param max_pruning: pruning level of the model for which ranking coefficient will be optimized
        :param random_seed: random seed, that will be set during ranking coefficients generation
        """
        self.num_generations = generations
        self.max_pruning = max_pruning
        self.train_steps = train_steps
        self.population_size = population_size
        self.num_samples = num_samples
        self.mutate_percent = mutate_percent
        self.scale_sigma = scale_sigma

        self.pruner = LeGRPruner(pruning_ctrl, target_model)
        init_filter_norms = self.pruner.init_filter_norms
        agent_hparams = {
            "num_generations": self.num_generations,
            "population_size": self.population_size,
            "num_samples": self.num_samples,
            "mutate_percent": self.mutate_percent,
            "sigma_scale": self.scale_sigma,
        }
        self.agent = EvolutionOptimizer(init_filter_norms, agent_hparams, random_seed)
        self.env = LeGREvolutionEnv(
            self.pruner,
            target_model,
            legr_init_args.train_loader,
            legr_init_args.val_loader,
            legr_init_args.train_steps_fn,
            legr_init_args.train_optimizer,
            legr_init_args.val_fn,
            legr_init_args.config,
            train_steps,
            max_pruning,
        )

    def train_global_ranking(self):
        """
        Training of ranking coefficients. During every generation:
        1. Environment (LeGREvolutionEnv) send reward and useful info from the previous generation to the
        agent (EvolutionOptimizer)
        2. Agent generates new action (considering this information)
        3. Environment makes step with this action, calculates and return current reward from this
         action and some useful info

         In the end, an optimal action from the agent is returned.
        :return: optimal ranking coefficients (action)
        """
        reward_list = []

        nncf_logger.info("Start training LeGR ranking coefficients...")

        generation_time = 0
        end = time.time()
        for episode in range(self.num_generations):
            state, info = self.env.reset()

            # Beginning of the episode
            done = 0
            reward = 0
            episode_reward = []
            self.agent.tell(state, reward, done, episode, info)

            while not done:
                action = self.agent.ask(episode)
                new_state, reward, done, info = self.env.step(action)
                self.agent.tell(state, reward, done, episode, info)

                state = new_state
                episode_reward.append(reward)
            generation_time = time.time() - end
            end = time.time()

            nncf_logger.info(
                f"Generation = {episode}, Reward = {episode_reward[0]:.3f}, Time = {generation_time:.3f} \n"
            )
            reward_list.append(episode_reward[0])
        self.env.reset()
        nncf_logger.info("Finished training LeGR ranking coefficients.")
        nncf_logger.info(f"Evolution algorithm rewards history = {reward_list}")

        best_ranking = self.agent.get_best_action()
        return best_ranking



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional, Tuple

import numpy as np
import onnx

from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.logging.logger import nncf_logger
from nncf.onnx.graph.metatypes import onnx_metatypes as om
from nncf.onnx.graph.metatypes.onnx_metatypes import ONNXDequantizeLinearMetatype
from nncf.onnx.graph.onnx_helper import get_tensor_value
from nncf.onnx.graph.transformations.commands import ONNXTargetPoint


def is_node_with_bias(node: NNCFNode) -> bool:
    """
    Checks if the node has a bias or not.

    :param node: The node to check.
    :return: Return `True` if `node` corresponds to the operation
        with bias (bias is added to the output tensor of that operation),
        `False` otherwise.
    """
    return node.layer_attributes.has_bias()


def get_bias_value(node_with_bias: NNCFNode, model: onnx.ModelProto) -> np.ndarray:
    """
    Returns the bias tensor for the biased node.

    :param node_with_bias: The node that corresponds to the operation with bias.
    :param model: The model that contains this operation.
    :return: The bias value that is applied to the output tensor of the node's operation.
    """
    assert node_with_bias.layer_attributes.has_bias()
    bias_name = node_with_bias.layer_attributes.bias_attrs["name"]
    return get_tensor_value(model, bias_name)


def get_input_edges_mapping(nncf_graph: NNCFGraph) -> Dict[str, Tuple[str, int]]:
    """
    Returns mapping between NNCFGraph input nodes and following by ONNX nodes with corresponding input port ids.

    :param nncf_graph: instance of NNCFGraph
    :return: A mapping of NNCF input node names and a tuple with the consumed node names and their input port ids.
    """
    input_edges_mapping = {}
    for input_node in nncf_graph.get_input_nodes():
        input_edges_mapping[input_node.node_name] = []
        for next_node in nncf_graph.get_next_nodes(input_node):
            for edge in nncf_graph.get_input_edges(next_node):
                if edge.from_node == input_node:
                    input_edges_mapping[input_node.node_name].append((next_node.node_name, edge.input_port_id))
    return input_edges_mapping


def get_input_edge(
    input_node_name: str,
    input_edges_mapping: Dict[str, Tuple[str, int]],
    node_mapping: Dict[str, onnx.NodeProto],
) -> str:
    """
    Returns input edge corresponding to the NNCF input node with the name input_node_name.

    :param input_node_name: Name of NNCF input node.
    :param input_edges_mapping: A mapping of NNCF input node names and
    a tuple with the consumed node names and their input port ids.
    :param node_mapping: Mapping of node names to the nodes.
    :return: Input edge name.
    """
    input_edges = set()
    for node_info in input_edges_mapping[input_node_name]:
        name, port_id = node_info
        node = node_mapping[name]
        input_edges.add(node.input[port_id])
    assert len(input_edges) == 1
    return input_edges.pop()


def is_any_weight_quantized(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
    """
    Returns True if any weight port id of node is quantized,
    False - if all weights are not quantized or the node can not have weight.

    :param node: NNCFNode.
    :param nncf_graph: NNCGraph.
    :return: True if any weight port id of node is quantized,
    False - if all weights are not quantized or the node can not have weight.
    """
    is_quantized_weight = False
    if node.layer_attributes.has_weight():
        for port_id in node.layer_attributes.weight_attrs:
            is_quantized_weight = is_quantized_weight or is_port_quantized(node, nncf_graph, port_id)
    return is_quantized_weight


def is_port_quantized(node: NNCFNode, nncf_graph: NNCFGraph, port_id: int) -> bool:
    """
    Returns True if a port_id is quantized - have ONNXDequantizeLinearMetatype as a parent node.

    :param node: NNCFNode.
    :param nncf_graph: NNCFGraph.
    :param port_id: Input port id of a node.
    :return: True if a port_id is quantized - have ONNXDequantizeLinearMetatype as a parent node.
    """
    input_nodes = [edge.from_node for edge in nncf_graph.get_input_edges(node)]
    if len(input_nodes) > port_id:
        weight_node = input_nodes[port_id]
        return weight_node.metatype == ONNXDequantizeLinearMetatype
    return False


def transpose_axis(shape: List[int], axis: int) -> int:
    """
    Returns transpose axis.

    :param shape: Tensor shape.
    :param axis: Axis before transpose (only positive).
    :return: Axis after transpose.
    """
    assert axis >= 0
    return range(len(shape) - 1, -1, -1)[axis]  # Iterate backward throug axis


def get_weight_quantization_axis(node: NNCFNode, port_id: int) -> int:
    """
    Returns weight tensor axis, along which quantizer parameters are calculated.

    :param node: NNCFNode, which has a weight on input port_id.
    :param port_id: Input port id on which there is a weight of a node.
    :return: Axis, along which quantizer parameters are calculated.
    """
    weight_channel_axis = node.metatype.weight_channel_axis
    if node.layer_attributes.has_node_attrs() and node.metatype == om.ONNXGemmMetatype:
        weight_shape = node.layer_attributes.weight_attrs[port_id]["shape"]
        weight_channel_axis %= len(weight_shape)  # Make axis positive
        if port_id == 0:
            weight_channel_axis -= 1
        if (
            port_id == 0
            and node.layer_attributes.node_attrs["transA"] == 1
            or port_id == 1
            and node.layer_attributes.node_attrs["transB"] == 1
        ):
            weight_channel_axis = transpose_axis(weight_shape, weight_channel_axis)
    return weight_channel_axis


def _get_activation_tensor_shape(
    nncf_graph: NNCFGraph, node: NNCFNode, target_point: ONNXTargetPoint
) -> Optional[Tuple[int, ...]]:
    """
    Returns shape of an activation tensor which is correspond to the target point and node.
    ONNX model can not have a shape of a edge, even after shape inference.
    Therefore, if there is no info regarding shape, None is returned.

    :param nncf_graph: NNCFGraph.
    :param node: NNCFNode.
    :param target_point: Determines from input or ouput of a node take a shape info.
    :return: None, if there is no shape info, otherwise - tensor shape.
    """
    if target_point.type == TargetType.PRE_LAYER_OPERATION:
        shape = nncf_graph.get_input_edges(node)[target_point.port_id].tensor_shape
    elif target_point.type == TargetType.POST_LAYER_OPERATION:
        shape = nncf_graph.get_output_edges(node)[target_point.port_id].tensor_shape
    else:
        raise NotImplementedError(f"Unsupported target point type {target_point.type}.")
    if not shape:  # ONNX model can not have a shape of a edge, even after shape inference.
        if target_point.type == TargetType.PRE_LAYER_OPERATION:
            nncf_logger.info(
                f"The shape of input edge of a node {node.node_name} is unkown. \
                    It could lead to inaccurate statistics collection."
            )
        elif target_point.type == TargetType.POST_LAYER_OPERATION:
            nncf_logger.info(
                f"The shape of output edge of a node {node.node_name} is unkown. \
                    It could lead to inaccurate statistics collection."
            )
        nncf_logger.info("Please consider to run pre-processing before quantization.")
        # TODO: add preprocessing tool for ONNX model.
        return None
    return shape


def get_quantized_tensor_shape(
    nncf_graph: NNCFGraph, node: NNCFNode, target_point: ONNXTargetPoint
) -> Optional[Tuple[int, ...]]:
    """
    Returns quantized tensor shape corresponding to a target point with a node if shape - info is existed.
    If there is no shape info - returns None.

    :param nncf_graph: NNCFGraph.
    :param node: NNCFNode.
    :param target_point: Target point indicates the quantizer place in the model graph.
    :return: Shape of a quantized tensor, if shape is existed. Otherwise - None.
    """
    if target_point.is_weight_target_point():
        return node.layer_attributes.weight_attrs[target_point.port_id]["shape"]
    return _get_activation_tensor_shape(nncf_graph, node, target_point)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from dataclasses import dataclass
from typing import Optional, Tuple, TypeVar

import numpy as np

from nncf.common.graph.graph import NNCFNode
from nncf.parameters import CompressWeightsMode

TWeightType = TypeVar("TWeightType")


@dataclass
class WeightCompressionConfig:
    """
    Configuration on how to compress (quantize) a specific weight.

    :param mode: Defines a mode for weight compression. Defaults to INT8_ASYM mode.
    :param group_size: Number of weights (e.g. 128) in the channel dimension that share quantization parameters (scale).
        The value -1 means no grouping. Defaults to -1.
    """

    mode: Optional[CompressWeightsMode] = CompressWeightsMode.INT8_ASYM
    group_size: Optional[int] = -1

    @property
    def num_bits(self):
        """
        :return: number of bits that is used for storing a single quantized value in the given mode.
        """
        return 8 if self.mode in [CompressWeightsMode.INT8_SYM, CompressWeightsMode.INT8_ASYM] else 4


@dataclass
class WeightCompressionParameters:
    """
    Weight compression parameters determine how and what weight should be compressed.

    :param weight_name: Unique weight name.
    :param node_with_weight: Node with weight in the NNCF graph.
    :param weight_port_id: Number of elements in the weight array.
    :param num_weights: Number of elements in the weight array.
    :param reduction_axes: Axes, along which to reduce (collect) different statistics (e.g. min, max).
    :param compression_config: Configuration of weight compression for the weight node.
    """

    weight_name: str
    node_with_weight: NNCFNode
    weight_port_id: int
    num_weights: np.uint64
    reduction_axes: Tuple[int, ...]
    compression_config = WeightCompressionConfig()

    def __post_init__(self):
        # Explicitly cast num_weights to avoid overflow on finding total number of weights.
        # The issue happens on Windows, because np.ndarray.size() returns np.int32 and sum of weights is more than 2^32.
        self.num_weights = np.uint64(self.num_weights)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nncf.common.graph.patterns import GraphPattern
from nncf.common.graph.patterns import HWFusedPatternNames
from nncf.common.utils.registry import Registry
from nncf.torch.graph.operator_metatypes import PTInputNoopMetatype
from nncf.torch.graph.pattern_operations import ARITHMETIC_OPERATIONS
from nncf.torch.graph.pattern_operations import ATOMIC_ACTIVATIONS_OPERATIONS
from nncf.torch.graph.pattern_operations import BATCH_NORMALIZATION_OPERATIONS
from nncf.torch.graph.pattern_operations import GROUP_NORMALIZATION_OPERATIONS
from nncf.torch.graph.pattern_operations import LINEAR_OPERATIONS
from nncf.torch.graph.pattern_operations import RELU_OPERATIONS

PT_HW_FUSED_PATTERNS = Registry("torch")

# ATOMIC OPERATIONS


@PT_HW_FUSED_PATTERNS.register(HWFusedPatternNames.L2_NORM)
def create_l2_norm_operations() -> GraphPattern:
    pattern = GraphPattern()

    outside_pattern_node = pattern.add_node(label="*OUTSIDE_PATTERN_NODE*", type=GraphPattern.NON_PATTERN_NODE_TYPE)
    pow_node = pattern.add_node(label="POW", type="pow")
    sum_node = pattern.add_node(label="SUM", type="sum")
    sqrt_node = pattern.add_node(label="SQRT", type="sqrt")
    add_node = pattern.add_node(label="ADD", type="__add__")
    div_node = pattern.add_node(label="DIV", type="div")
    mul_node = pattern.add_node(label="MUL", type="__rmul__")

    pattern.add_edge(outside_pattern_node, pow_node)
    pattern.add_edge(pow_node, sum_node)
    pattern.add_edge(sum_node, sqrt_node)
    pattern.add_edge(sqrt_node, add_node)
    pattern.add_edge(add_node, div_node)
    pattern.add_edge(div_node, mul_node)
    pattern.add_edge(outside_pattern_node, div_node)
    return pattern


# COMBINATIONS


@PT_HW_FUSED_PATTERNS.register(HWFusedPatternNames.SHIFT_SCALE)
def create_shift_scale() -> GraphPattern:
    pattern = GraphPattern()
    add_node = pattern.add_node(label="ADD, SUB", type=["__add__", "__sub__"])
    truediv_node = pattern.add_node(label="MUL, DIV", type=["__mul__", "__truediv__"])
    pattern.add_edge(add_node, truediv_node)
    return pattern


@PT_HW_FUSED_PATTERNS.register(HWFusedPatternNames.INPUT_SHIFT_SCALE)
def create_input_shift_scale() -> GraphPattern:
    pattern = GraphPattern()
    pattern.add_node(**{GraphPattern.LABEL_ATTR: "MODEL_INPUT", GraphPattern.METATYPE_ATTR: PTInputNoopMetatype})
    shift_scale = create_shift_scale()
    pattern.join_patterns(shift_scale)
    return pattern


@PT_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_ARITHMETIC)
def create_linear_arithmetic_operations() -> GraphPattern:
    linear = linear_operations()
    arithmetic = arithmetic_operations()
    linear.join_patterns(arithmetic)
    return linear


@PT_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_SHIFT_SCALE)
def create_linear_shift_scale() -> GraphPattern:
    linear = linear_operations()
    shift_scale = create_shift_scale()
    linear.join_patterns(shift_scale)
    return linear


@PT_HW_FUSED_PATTERNS.register(HWFusedPatternNames.BATCH_NORM_ACTIVATIONS)
def create_batch_norm_activations_operations() -> GraphPattern:
    batch_norm = batch_norm_operations()
    activations = activation_operations()
    batch_norm.join_patterns(activations)
    return batch_norm


@PT_HW_FUSED_PATTERNS.register(HWFusedPatternNames.ACTIVATIONS_BATCH_NORM)
def create_activations_batch_norm_operations() -> GraphPattern:
    batch_norm = batch_norm_operations()
    activations = activation_operations()
    activations.join_patterns(batch_norm)
    return activations


@PT_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_BATCH_NORM)
def create_linear_batch_norm_operations() -> GraphPattern:
    linear = linear_operations()
    batch_norm = batch_norm_operations()
    linear.join_patterns(batch_norm)
    return linear


@PT_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_ACTIVATIONS)
def create_linear_activation_operations() -> GraphPattern:
    linear = linear_operations()
    activation = activation_operations()
    linear.join_patterns(activation)
    return linear


@PT_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_BATCH_NORM_ACTIVATIONS)
def create_linear_batch_norm_activation_operations() -> GraphPattern:
    linear_bn = create_linear_batch_norm_operations()
    activations = activation_operations()
    linear_bn.join_patterns(activations)
    return linear_bn


@PT_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_ACTIVATIONS_BATCH_NORM)
def create_linear_activation_batch_norm_activations() -> GraphPattern:
    linear_act = create_linear_activation_operations()
    batch_norm = batch_norm_operations()
    linear_act.join_patterns(batch_norm)
    return linear_act


@PT_HW_FUSED_PATTERNS.register(HWFusedPatternNames.ARITHMETIC_BATCH_NORM)
def create_arithmetic_batch_norm_operations() -> GraphPattern:
    arithmetic = arithmetic_operations()
    batch_norm = batch_norm_operations()
    arithmetic.join_patterns(batch_norm)
    return arithmetic


@PT_HW_FUSED_PATTERNS.register(HWFusedPatternNames.ARITHMETIC_ACTIVATIONS)
def create_arithmetic_activations_operations() -> GraphPattern:
    arithmetic = arithmetic_operations()
    activation = activation_operations()
    arithmetic.join_patterns(activation)
    return arithmetic


@PT_HW_FUSED_PATTERNS.register(HWFusedPatternNames.ARITHMETIC_BATCH_NORM_ACTIVATIONS)
def create_arithmetic_batch_norm_activations_operations() -> GraphPattern:
    arithmetic_bn = create_arithmetic_batch_norm_operations()
    activation = activation_operations()
    arithmetic_bn.join_patterns(activation)
    return arithmetic_bn


@PT_HW_FUSED_PATTERNS.register(HWFusedPatternNames.ARITHMETIC_ACTIVATIONS_BATCH_NORM)
def create_arithmetic_activations_batch_norm_operations() -> GraphPattern:
    arithmetic_act = create_arithmetic_activations_operations()
    batch_norm = batch_norm_operations()
    arithmetic_act.join_patterns(batch_norm)
    return arithmetic_act


@PT_HW_FUSED_PATTERNS.register(HWFusedPatternNames.GROUP_NORM_RELU)
def create_group_norm_relu_operations() -> GraphPattern:
    group_norm = GraphPattern()
    group_norm.add_node(**GROUP_NORMALIZATION_OPERATIONS)
    relu = GraphPattern()
    relu.add_node(**RELU_OPERATIONS)
    group_norm.join_patterns(relu)
    return group_norm


@PT_HW_FUSED_PATTERNS.register(HWFusedPatternNames.LINEAR_CONST_MULTIPLY)
def create_linear_const_multiply() -> GraphPattern:
    pattern = GraphPattern()
    linear_node = pattern.add_node(label="linear", type="linear")
    mul_node = pattern.add_node(label="MUL", type="__mul__")
    pattern.add_edge(linear_node, mul_node)

    return pattern


def linear_operations() -> GraphPattern:
    pattern = GraphPattern()
    pattern.add_node(**LINEAR_OPERATIONS)
    return pattern


def arithmetic_operations() -> GraphPattern:
    pattern = GraphPattern()
    pattern.add_node(**ARITHMETIC_OPERATIONS)
    return pattern


def batch_norm_operations() -> GraphPattern:
    pattern = GraphPattern()
    pattern.add_node(**BATCH_NORMALIZATION_OPERATIONS)
    return pattern


def activation_operations() -> GraphPattern:
    atomic_activations = GraphPattern()
    atomic_activations.add_node(**ATOMIC_ACTIVATIONS_OPERATIONS)
    swish = create_swish_act()
    h_sigmoid = create_h_sigmoid_act()
    h_swish = create_h_swish_act()

    pattern = GraphPattern()
    pattern.add_pattern_alternative(atomic_activations)
    pattern.add_pattern_alternative(swish)
    pattern.add_pattern_alternative(h_swish)
    pattern.add_pattern_alternative(h_sigmoid)
    return pattern


def create_swish_act() -> GraphPattern:
    pattern = GraphPattern()
    input_pattern_node = pattern.add_node(label="*INPUT_NODE*", type=GraphPattern.NON_PATTERN_NODE_TYPE)
    sigmoid_node = pattern.add_node(label="SIGMOID", type="sigmoid")
    mul_node = pattern.add_node(label="MUL", type="__mul__")

    pattern.add_edge(input_pattern_node, sigmoid_node)
    pattern.add_edge(sigmoid_node, mul_node)
    pattern.add_edge(input_pattern_node, mul_node)
    return pattern


def create_h_swish_act() -> GraphPattern:
    main_pattern = GraphPattern()

    # Mul -> Div version
    pattern = GraphPattern()
    input_pattern_node = pattern.add_node(label="*INPUT_NODE*", type=GraphPattern.NON_PATTERN_NODE_TYPE)
    add_node = pattern.add_node(label="ADD", type="__add__")
    hardtanh_node = pattern.add_node(label="HARDTANH", type="hardtanh")
    truediv_node = pattern.add_node(label="DIV", type="__truediv__")
    mul_node = pattern.add_node(label="MUL", type="__mul__")

    pattern.add_edge(input_pattern_node, add_node)
    pattern.add_edge(input_pattern_node, mul_node)
    pattern.add_edge(add_node, hardtanh_node)
    pattern.add_edge(hardtanh_node, truediv_node)
    pattern.add_edge(truediv_node, mul_node)
    main_pattern.add_pattern_alternative(pattern)

    # Div -> Mul version
    pattern = GraphPattern()
    input_pattern_node = pattern.add_node(label="*INPUT_NODE*", type=GraphPattern.NON_PATTERN_NODE_TYPE)
    add_node = pattern.add_node(label="ADD", type="__add__")
    hardtanh_node = pattern.add_node(label="HARDTANH", type="hardtanh")
    mul_node = pattern.add_node(label="MUL", type="__mul__")
    truediv_node = pattern.add_node(label="DIV", type="__truediv__")

    pattern.add_edge(input_pattern_node, add_node)
    pattern.add_edge(input_pattern_node, mul_node)
    pattern.add_edge(add_node, hardtanh_node)
    pattern.add_edge(hardtanh_node, mul_node)
    pattern.add_edge(mul_node, truediv_node)
    main_pattern.add_pattern_alternative(pattern)

    # ReLU6 version - Mul -> Div
    pattern = GraphPattern()
    input_pattern_node = pattern.add_node(label="*INPUT_NODE*", type=GraphPattern.NON_PATTERN_NODE_TYPE)
    add_node = pattern.add_node(label="ADD", type="__add__")
    relu6_node = pattern.add_node(label="RELU6", type="relu6")
    mul_node = pattern.add_node(label="MUL", type="__mul__")
    truediv_node = pattern.add_node(label="DIV", type="__truediv__")

    pattern.add_edge(input_pattern_node, add_node)
    pattern.add_edge(input_pattern_node, mul_node)
    pattern.add_edge(add_node, relu6_node)
    pattern.add_edge(hardtanh_node, mul_node)
    pattern.add_edge(mul_node, truediv_node)
    main_pattern.add_pattern_alternative(pattern)

    # ReLU6 version - Div -> Mul
    pattern = GraphPattern()
    input_pattern_node = pattern.add_node(label="*INPUT_NODE*", type=GraphPattern.NON_PATTERN_NODE_TYPE)
    add_node = pattern.add_node(label="ADD", type="__add__")
    relu6_node = pattern.add_node(label="RELU6", type="relu6")
    truediv_node = pattern.add_node(label="DIV", type="__truediv__")
    mul_node = pattern.add_node(label="MUL", type="__mul__")

    pattern.add_edge(input_pattern_node, add_node)
    pattern.add_edge(input_pattern_node, mul_node)
    pattern.add_edge(add_node, relu6_node)
    pattern.add_edge(relu6_node, truediv_node)
    pattern.add_edge(truediv_node, mul_node)

    main_pattern.add_pattern_alternative(pattern)

    return main_pattern


def create_h_sigmoid_act() -> GraphPattern:
    main_pattern = GraphPattern()

    # ReLU version:
    pattern = GraphPattern()

    input_pattern_node = pattern.add_node(label="*INPUT_NODE*", type=GraphPattern.NON_PATTERN_NODE_TYPE)
    add_node = pattern.add_node(label="ADD", type="__add__")
    hardtanh_node = pattern.add_node(label="HARDTANH", type="hardtanh")
    truediv_node = pattern.add_node(label="DIV", type="__truediv__")

    pattern.add_edge(input_pattern_node, add_node)
    pattern.add_edge(add_node, hardtanh_node)
    pattern.add_edge(hardtanh_node, truediv_node)

    main_pattern.add_pattern_alternative(pattern)

    # ReLU6 version
    pattern = GraphPattern()

    input_pattern_node = pattern.add_node(label="*INPUT_NODE*", type=GraphPattern.NON_PATTERN_NODE_TYPE)
    add_node = pattern.add_node(label="ADD", type="__add__")
    relu6_node = pattern.add_node(label="RELU6", type="relu6")
    truediv_node = pattern.add_node(label="DIV", type="__truediv__")

    pattern.add_edge(input_pattern_node, add_node)
    pattern.add_edge(add_node, relu6_node)
    pattern.add_edge(hardtanh_node, truediv_node)

    main_pattern.add_pattern_alternative(pattern)

    return main_pattern



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Implementations of training loops to be used for accuracy aware training.
"""
import pathlib
from abc import ABC
from abc import abstractmethod
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple, TypeVar, Union, cast

import numpy as np
from scipy.interpolate import interp1d  # type: ignore

import nncf
from nncf.api.compression import CompressionAlgorithmController
from nncf.common.accuracy_aware_training.runner import BaseAdaptiveCompressionLevelTrainingRunner
from nncf.common.accuracy_aware_training.runner import TrainingRunner
from nncf.common.accuracy_aware_training.runner_factory import AdaptiveCompressionLevelTrainingRunnerCreator
from nncf.common.accuracy_aware_training.runner_factory import EarlyExitTrainingRunnerCreator
from nncf.common.accuracy_aware_training.statistics import TrainingLoopStatistics
from nncf.common.composite_compression import CompositeCompressionAlgorithmController
from nncf.common.logging import nncf_logger
from nncf.common.utils.api_marker import api
from nncf.common.utils.registry import Registry
from nncf.config.config import NNCFConfig
from nncf.config.extractors import extract_accuracy_aware_training_params

TModel = TypeVar("TModel")
TensorboardWriterType = TypeVar("TensorboardWriterType")
OptimizerType = TypeVar("OptimizerType")
LRSchedulerType = TypeVar("LRSchedulerType")
Checkpoint = TypeVar("Checkpoint")
ADAPTIVE_COMPRESSION_CONTROLLERS = Registry("adaptive_compression_controllers")


@api()
class TrainingLoop(ABC):
    """
    The training loop object that launches the training process via the `run` method.
    """

    @abstractmethod
    def run(
        self,
        model: TModel,
        train_epoch_fn: Callable[
            [
                CompressionAlgorithmController,
                TModel,
                Optional[int],
                Optional[OptimizerType],
                Optional[LRSchedulerType],
            ],
            float,
        ],
        validate_fn: Callable[[TModel], float],
        configure_optimizers_fn: Optional[Callable[[], Tuple[OptimizerType, LRSchedulerType]]] = None,
        dump_checkpoint_fn: Optional[
            Callable[
                [
                    TModel,
                    CompressionAlgorithmController,
                    TrainingRunner,
                    Union[str, pathlib.Path],
                ],
                None,
            ]
        ] = None,
        load_checkpoint_fn: Optional[Callable[[Checkpoint, Union[str, pathlib.Path]], None]] = None,
        early_stopping_fn: Optional[Callable[[float], bool]] = None,
        tensorboard_writer: Optional[TensorboardWriterType] = None,
        log_dir: Union[pathlib.Path, str] = None,
        update_learning_rate_fn: Optional[Callable[[], None]] = None,
    ) -> TModel:
        """
        Implements the custom logic to run a training loop for model fine-tuning by using the provided
        `train_epoch_fn`, `validate_fn` and `configure_optimizers_fn` methods.

        :param model: The model instance before fine-tuning
        :param train_epoch_fn: a callback to fine-tune the model for a single epoch
        :param validate_fn: a callback to evaluate the model on the validation dataset
        :param configure_optimizers_fn: a callback to instantiate an optimizer and a learning rate scheduler
        :param dump_checkpoint_fn: a callback to dump a checkpoint
        :param load_checkpoint_fn: a callback to load a checkpoint
        :param early_stopping_fn: a callback to check for an early stopping condition
        :param tensorboard_writer: The tensorboard object to be used for logging.
        :param log_dir: The path to be used for logging and checkpoint saving.
        :param update_learning_rate_fn: The callback to update the learning rate after each epoch
          of the training loop.
        :return: The fine-tuned model.
        """

    @property
    @abstractmethod
    def statistics(self) -> TrainingLoopStatistics:
        """
        Returns statistics of the compressed model.
        """


@api()
class BaseEarlyExitCompressionTrainingLoop(TrainingLoop, ABC):
    """
    Base class to generalize functionality of derived training loop classes.
    """

    def __init__(self, compression_controller: CompressionAlgorithmController) -> None:
        self.runner: BaseAdaptiveCompressionLevelTrainingRunner
        self.compression_controller: CompressionAlgorithmController = compression_controller
        self._current_compression_rate: float

    def run(
        self,
        model: TModel,
        train_epoch_fn: Callable[
            [
                CompressionAlgorithmController,
                TModel,
                Optional[int],
                Optional[OptimizerType],
                Optional[LRSchedulerType],
            ],
            float,
        ],
        validate_fn: Callable[[TModel], float],
        configure_optimizers_fn: Optional[Callable[[], Tuple[OptimizerType, LRSchedulerType]]] = None,
        dump_checkpoint_fn: Optional[
            Callable[
                [
                    TModel,
                    CompressionAlgorithmController,
                    TrainingRunner,
                    Union[str, pathlib.Path],
                ],
                None,
            ]
        ] = None,
        load_checkpoint_fn: Optional[Callable[[Checkpoint, Union[str, pathlib.Path]], None]] = None,
        early_stopping_fn: Optional[Callable[[float], bool]] = None,
        tensorboard_writer: Optional[TensorboardWriterType] = None,
        log_dir: Optional[Union[pathlib.Path, str]] = None,
        update_learning_rate_fn: Optional[Callable[[], None]] = None,
    ) -> TModel:
        self.runner.initialize_training_loop_fns(
            train_epoch_fn,
            validate_fn,
            configure_optimizers_fn,
            dump_checkpoint_fn,  # type: ignore
            load_checkpoint_fn,
            early_stopping_fn,
            update_learning_rate_fn,
        )
        self.runner.initialize_logging(log_dir, tensorboard_writer)
        return self._run_early_exit_training_loop(model)

    def _run_early_exit_training_loop(self, model: TModel) -> TModel:
        uncompressed_model_accuracy = self.runner.uncompressed_model_accuracy
        self.runner.calculate_minimal_tolerable_accuracy(uncompressed_model_accuracy)

        self.runner.configure_optimizers()

        self.runner.validate(model)
        self._current_compression_rate = self.compression_controller.compression_rate
        self.runner.dump_statistics(model, self.compression_controller)

        nncf_logger.info("Initialization step results:")
        self.log_accuracy_statistics()

        if self._accuracy_criterion_satisfied():
            nncf_logger.info("\nReached the accuracy criteria after the initialization step.\n")
            return model
        maximal_total_epochs = cast(int, self.runner.maximal_total_epochs)
        for epoch in range(1, maximal_total_epochs + 1):
            self.runner.train_epoch(model, self.compression_controller)
            self.runner.validate(model)
            self._current_compression_rate = self.compression_controller.compression_rate
            self.runner.dump_statistics(model, self.compression_controller)
            if self._accuracy_criterion_satisfied():
                nncf_logger.info(f"Reached the accuracy criteria after epoch {epoch}.")
                self.log_accuracy_statistics()
                break

            nncf_logger.info(f"Epoch {epoch} results:")
            self.log_accuracy_statistics()

            if self.runner.stop_training(self.compression_controller):
                nncf_logger.info("Training stopped - early stopping criterion satisfied")
                break

            self.runner.update_learning_rate()

        self._current_compression_rate = self.runner.load_best_checkpoint(model)
        return model

    def log_accuracy_statistics(self) -> None:
        for log_str in self.statistics.to_str().split("\n"):
            nncf_logger.info(log_str)

    @property
    def statistics(self) -> TrainingLoopStatistics:
        compression_rate = self._current_compression_rate or 0.0
        compressed_accuracy = self.runner.current_val_metric_value
        uncompressed_accuracy = self.runner.uncompressed_model_accuracy
        accuracy_drop = self._calculate_accuracy_drop(uncompressed_accuracy, compressed_accuracy)
        relative_accuracy_drop = self._calculate_rel_accuracy_drop(uncompressed_accuracy, compressed_accuracy)
        accuracy_budget = self._calculate_accuracy_budget(self.runner.minimal_tolerable_accuracy, compressed_accuracy)
        stats = TrainingLoopStatistics(
            uncompressed_accuracy,
            compression_rate,
            compressed_accuracy,
            accuracy_drop,
            relative_accuracy_drop,
            accuracy_budget,
        )
        return stats

    @staticmethod
    def _calculate_accuracy_drop(uncompressed_model_accuracy: float, compressed_model_accuracy: float) -> float:
        return uncompressed_model_accuracy - compressed_model_accuracy

    @staticmethod
    def _calculate_accuracy_budget(minimal_tolerable_accuracy: float, compressed_model_accuracy: float) -> float:
        return compressed_model_accuracy - minimal_tolerable_accuracy

    @staticmethod
    def _calculate_rel_accuracy_drop(uncompressed_model_accuracy: float, compressed_model_accuracy: float) -> float:
        try:
            rel_accuracy_drop = 100 * (1.0 - compressed_model_accuracy / uncompressed_model_accuracy)
        except ZeroDivisionError:
            rel_accuracy_drop = 0

        return rel_accuracy_drop

    def _accuracy_criterion_satisfied(self) -> bool:
        accuracy_budget = self._calculate_accuracy_budget(
            self.runner.minimal_tolerable_accuracy, self.runner.current_val_metric_value
        )
        return accuracy_budget >= 0 and self.runner.is_model_fully_compressed(self.compression_controller)


@api()
class EarlyExitCompressionTrainingLoop(BaseEarlyExitCompressionTrainingLoop):
    """
    Training loop that does not modify compression parameters and exits as soon as (and if) the accuracy drop criterion
    is reached.

    :param nncf_config: The configuration object.
    :type nncf_config: nncf.NNCFConfig
    :param compression_controller: The controller for the compression algorithm that is currently applied to the model
        to be trained.
    :param uncompressed_model_accuracy: The uncompressed model accuracy, measured outside of this training loop to
        serve as the point of reference for fine-tuning the compressed model.
    :param lr_updates_needed:
    :param verbose: Whether to post additional data to TensorBoard.
    :param dump_checkpoints: If true, will dump all checkpoints obtained during the training process, otherwise will
      only keep the best checkpoint (accuracy-wise).
    """

    def __init__(
        self,
        nncf_config: NNCFConfig,
        compression_controller: CompressionAlgorithmController,
        uncompressed_model_accuracy: float,
        lr_updates_needed: bool = True,
        verbose: bool = True,
        dump_checkpoints: bool = True,
    ) -> None:
        super().__init__(compression_controller)
        accuracy_aware_training_params = extract_accuracy_aware_training_params(nncf_config)
        runner_factory = EarlyExitTrainingRunnerCreator(
            accuracy_aware_training_params,  # type: ignore
            compression_controller,
            uncompressed_model_accuracy,
            verbose,
            dump_checkpoints,
            lr_updates_needed,
        )

        self.runner = runner_factory.create_training_loop()  # type: ignore


@api()
class AdaptiveCompressionTrainingLoop(BaseEarlyExitCompressionTrainingLoop):
    """
    A training loop that automatically adjusts compression rate to reach maximum compression within accuracy budget.

    :param nncf_config: The configuration object.
    :type nncf_config: nncf.NNCFConfig
    :param compression_controller: The controller for the compression algorithm that is currently applied to the model
        to be trained.
    :param uncompressed_model_accuracy: The uncompressed model accuracy, measured outside of this training loop to
        serve as the point of reference for fine-tuning the compressed model.
    :param lr_updates_needed:
    :param verbose: Whether to post additional data to TensorBoard.
    :param minimal_compression_rate: Sets the minimal compression rate to be considered during the training loop.
    :param maximal_compression_rate: Sets the maximal compression rate to be considered during the training loop.
    :param dump_checkpoints: If true, will dump all checkpoints obtained during the training process, otherwise will
      only keep the best checkpoint (accuracy-wise).
    """

    def __init__(
        self,
        nncf_config: NNCFConfig,
        compression_controller: CompressionAlgorithmController,
        uncompressed_model_accuracy: float,
        lr_updates_needed: bool = True,
        verbose: bool = True,
        minimal_compression_rate: float = 0.0,
        maximal_compression_rate: float = 0.95,
        dump_checkpoints: bool = True,
    ) -> None:
        super().__init__(compression_controller)
        self.adaptive_controller = self._get_adaptive_compression_ctrl(compression_controller)
        if self.adaptive_controller is None:
            raise nncf.InternalError(
                "No compression algorithm supported by the accuracy-aware training "
                "runner was specified in the config"
            )

        maximal_compression_rate = min(maximal_compression_rate, self.adaptive_controller.maximal_compression_rate)

        accuracy_aware_training_params = extract_accuracy_aware_training_params(nncf_config)
        runner_factory = AdaptiveCompressionLevelTrainingRunnerCreator(
            accuracy_aware_training_params,  # type: ignore
            self.adaptive_controller,
            uncompressed_model_accuracy,
            verbose,
            dump_checkpoints,
            lr_updates_needed,
            minimal_compression_rate,
            maximal_compression_rate,
        )
        self.runner = runner_factory.create_training_loop()
        self.runner.adaptive_controller = self.adaptive_controller

    def _get_adaptive_compression_ctrl(
        self, compression_controller: CompressionAlgorithmController
    ) -> CompressionAlgorithmController:
        def _adaptive_compression_controllers() -> Dict[str, CompressionAlgorithmController]:
            def remove_registry_prefix(algo_name: str) -> str:
                for prefix in ("pt_", "tf_"):
                    if algo_name.startswith(prefix):
                        return algo_name[len(prefix) :]
                raise nncf.ValidationError(
                    "Compression algorithm names in the adaptive controllers "
                    'registry should be prefixed with "pt_" or "tf_" depending on the '
                    "backend framework"
                )

            return {
                remove_registry_prefix(algo_name): controller_cls
                for algo_name, controller_cls in ADAPTIVE_COMPRESSION_CONTROLLERS.registry_dict.items()
            }

        adaptive_compression_controllers = _adaptive_compression_controllers()

        if isinstance(compression_controller, CompositeCompressionAlgorithmController):
            for controller in compression_controller.child_ctrls:
                for ctrl_type in adaptive_compression_controllers.values():
                    if isinstance(controller, ctrl_type):  # type: ignore
                        return controller
        elif (
            isinstance(compression_controller, CompressionAlgorithmController)
            and compression_controller.name in adaptive_compression_controllers
        ):
            return compression_controller

        raise nncf.InternalError(
            "No compression algorithm that supports adaptive compression accuracy-aware training was specified"
        )

    def run(
        self,
        model: TModel,
        train_epoch_fn: Callable[
            [
                CompressionAlgorithmController,
                TModel,
                Optional[int],
                Optional[OptimizerType],
                Optional[LRSchedulerType],
            ],
            float,
        ],
        validate_fn: Callable[[TModel], float],
        configure_optimizers_fn: Optional[Callable[[], Tuple[OptimizerType, LRSchedulerType]]] = None,
        dump_checkpoint_fn: Optional[
            Callable[
                [
                    TModel,
                    CompressionAlgorithmController,
                    TrainingRunner,
                    Union[str, pathlib.Path],
                ],
                None,
            ]
        ] = None,
        load_checkpoint_fn: Optional[Callable[[Checkpoint, Union[str, pathlib.Path]], None]] = None,
        early_stopping_fn: Optional[Callable[[float], bool]] = None,
        tensorboard_writer: Optional[TensorboardWriterType] = None,
        log_dir: Optional[Union[pathlib.Path, str]] = None,
        update_learning_rate_fn: Optional[Callable[[], None]] = None,
    ) -> TModel:
        self.runner.initialize_training_loop_fns(
            train_epoch_fn,
            validate_fn,
            configure_optimizers_fn,
            dump_checkpoint_fn,  # type: ignore
            load_checkpoint_fn,
            early_stopping_fn,
            update_learning_rate_fn,
        )
        self.runner.initialize_logging(log_dir, tensorboard_writer)
        model = self._run_initial_training_phase(model)
        self.runner.reset_training()
        self.runner.validate(model)

        nncf_logger.info("Search for the optimal compression rate started.")
        accuracy_budget = self._calculate_accuracy_budget(
            self.runner.minimal_tolerable_accuracy, self.runner.best_val_metric_value
        )
        self.runner.add_tensorboard_scalar(
            "val/accuracy_aware/accuracy_budget",
            accuracy_budget,
            self.runner.cumulative_epoch_count,
        )
        self.runner.add_tensorboard_scalar(
            "compression/accuracy_aware/target_compression_rate",
            self.runner.compression_rate_target,
            self.runner.cumulative_epoch_count,
        )

        self.compression_controller.disable_scheduler()
        force_updating_target_compression_rate = True
        while (
            self.runner.compression_rate_step >= self.runner.minimal_compression_rate_step
            and self.runner.cumulative_epoch_count < self.runner.maximal_total_epochs
        ):
            prev_compression_rate_target = self.runner.compression_rate_target
            prev_compression_rate_step = self.runner.compression_rate_step
            was_compression_rate_changed = self._update_target_compression_rate(
                self.runner, force_updating_target_compression_rate
            )

            self.log_accuracy_statistics()

            if was_compression_rate_changed:
                nncf_logger.info(
                    f"Target compression rate value changed: {prev_compression_rate_target:.3f} -> "
                    f"{self.runner.compression_rate_target:.3f}"
                )
                if prev_compression_rate_step == self.runner.compression_rate_step:
                    nncf_logger.info(
                        f"Compression rate step value is kept unchanged: {self.runner.compression_rate_step:.3f}"
                    )
                else:
                    nncf_logger.info(
                        f"The compression rate step value was changed {prev_compression_rate_step:.3f} -> "
                        f"{self.runner.compression_rate_step:.3f}"
                    )
                if self.runner.compression_rate_target < self.runner.minimal_compression_rate:
                    nncf_logger.warning("Cannot produce a compressed model with a specified minimal tolerable accuracy")
                    break
                if self.runner.compression_rate_target > self.runner.maximal_compression_rate:
                    self.runner.compression_rate_target = self.runner.maximal_compression_rate
                    nncf_logger.info(
                        f"Reached maximal possible compression rate: {self.runner.maximal_compression_rate}"
                    )
                    break

                force_updating_target_compression_rate = False

                self.runner.load_best_checkpoint(model)
                # set compression rate for model
                self.adaptive_controller.compression_rate = self.runner.compression_rate_target
                self._current_compression_rate = self.runner.compression_rate_target
                self.runner.reset_training()

                # workaround for compression statistics
                self.adaptive_controller.scheduler.current_sparsity_level = self.runner.compression_rate_target  # type: ignore
                self.adaptive_controller.scheduler.current_pruning_level = self.runner.compression_rate_target  # type: ignore
                self.adaptive_controller.scheduler.target_level = self.runner.compression_rate_target  # type: ignore

                self.runner.add_tensorboard_scalar(
                    "compression/accuracy_aware/target_compression_rate",
                    self.runner.compression_rate_target,
                    self.runner.cumulative_epoch_count,
                )
                self.runner.add_tensorboard_scalar(
                    "compression/accuracy_aware/compression_rate_step",
                    self.runner.compression_rate_step,
                    self.runner.cumulative_epoch_count,
                )
            else:
                nncf_logger.info(f"Current target compression rate value: {self.runner.compression_rate_target:.3f}")
                nncf_logger.info(f"Current compression rate step value: {self.runner.compression_rate_target:.3f}")

            self.runner.train_epoch(model, self.compression_controller)
            compressed_model_accuracy = self.runner.validate(model)
            self.runner.dump_statistics(model, self.compression_controller)
            accuracy_budget = self._calculate_accuracy_budget(
                self.runner.minimal_tolerable_accuracy, compressed_model_accuracy
            )
            self.runner.add_tensorboard_scalar(
                "val/accuracy_aware/accuracy_budget",
                accuracy_budget,
                self.runner.cumulative_epoch_count,
            )
            if self.runner.stop_training(self.compression_controller):
                nncf_logger.info("Training stopped - early stopping criterion satisfied.")
                force_updating_target_compression_rate = True

            self.runner.update_learning_rate()

        self._current_compression_rate = self.runner.load_best_checkpoint(model)
        compressed_model_accuracy = self.runner.validate(model)
        nncf_logger.info(
            f"Compression rate for the final compressed model: {self._current_compression_rate}, "
            f"accuracy: {compressed_model_accuracy} "
            f"(vs. uncompressed model accuracy: {self.runner.uncompressed_model_accuracy})"
        )
        return model

    def _run_initial_training_phase(self, model: TModel) -> TModel:
        nncf_logger.info("Initial training phase started...")

        maximal_total_epochs = self.runner.maximal_total_epochs
        self.runner.maximal_total_epochs = self.runner.initial_training_phase_epochs
        model = self._run_early_exit_training_loop(model)
        self.runner.maximal_total_epochs = maximal_total_epochs

        nncf_logger.info("Initial training phase finished.")
        return model

    def _update_target_compression_rate(
        self,
        runner: BaseAdaptiveCompressionLevelTrainingRunner,
        force_update: bool = False,
    ) -> bool:
        best_accuracy_budget = runner.best_val_metric_value - runner.minimal_tolerable_accuracy
        nncf_logger.info(
            f"Training epoch count: {runner.training_epoch_count}, patience epochs: {runner.patience_epochs}"
        )
        if runner.training_epoch_count >= runner.patience_epochs or best_accuracy_budget >= 0.0 or force_update:
            runner.compression_rate_target += self._determine_compression_rate_step_value(runner)
            runner.was_compression_increased_on_prev_step = 1.0 if best_accuracy_budget >= 0.0 else -1.0
            return True
        return False

    def _determine_compression_rate_step_value(
        self,
        runner: BaseAdaptiveCompressionLevelTrainingRunner,
        stepping_mode: str = "uniform_decrease",
        **kwargs: Union[int, float],
    ) -> float:
        if stepping_mode == "uniform_decrease":
            compression_step_updater = self._uniform_decrease_compression_step_update
        elif stepping_mode == "interpolate":
            compression_step_updater = partial(
                self._interpolate_compression_step_update,
                current_compression_rate=runner.compression_rate_target,
            )
        else:
            raise ValueError("Wrong stepping mode to determine compression rate step value provided")
        return compression_step_updater(runner, **kwargs)

    @staticmethod
    def _uniform_decrease_compression_step_update(
        runner: BaseAdaptiveCompressionLevelTrainingRunner,
    ) -> float:
        best_accuracy_budget = runner.best_val_metric_value - runner.minimal_tolerable_accuracy
        best_accuracy_budget_sign = 1.0 if best_accuracy_budget >= 0.0 else -1.0
        # if we don't fit the accuracy budget now and before we did fit or vice versa, we reduce the compression rate
        # step and the learning rate
        if (
            runner.was_compression_increased_on_prev_step is not None
            and runner.was_compression_increased_on_prev_step != best_accuracy_budget_sign
        ):
            runner.compression_rate_step *= runner.compression_rate_step_reduction_factor
            runner.base_lr_reduction_factor_during_search *= runner.lr_reduction_factor
        # if we don't fit the accuracy budget, then we decrease the compression rate, and if otherwise we increase it
        compression_step_update = best_accuracy_budget_sign * runner.compression_rate_step
        return compression_step_update

    @staticmethod
    def _interpolate_compression_step_update(
        runner: BaseAdaptiveCompressionLevelTrainingRunner,
        current_compression_rate: float,
        num_curve_pts: int = 1000,
        full_compression_factor: int = 20,
        minimal_compression_rate: float = 0.0,
        maximal_compression_rate: float = 1.0,
    ) -> float:
        training_history = runner.compressed_training_history
        nncf_logger.info(f"Compressed training history: {training_history}")
        training_history[minimal_compression_rate] = runner.maximal_accuracy_drop  # type: ignore
        training_history[maximal_compression_rate] = -full_compression_factor * runner.maximal_accuracy_drop  # type: ignore
        compression_rates, evaluated_acc_budgets = cast(List[float], training_history.keys()), cast(
            List[float], training_history.values()
        )
        interp_kind = "linear" if len(compression_rates) < 4 else "cubic"
        acc_budget_vs_comp_rate_curve = interp1d(compression_rates, evaluated_acc_budgets, kind=interp_kind)
        rate_interval = np.linspace(
            minimal_compression_rate,
            maximal_compression_rate,
            num=num_curve_pts,
            endpoint=True,
        )
        acc_budget_values = acc_budget_vs_comp_rate_curve(rate_interval)
        target_compression_rate: float = rate_interval[np.argmin(np.abs(acc_budget_values))]
        nncf_logger.info(
            f"Predicted compression rate: {target_compression_rate}, "
            f"current compression rate: {current_compression_rate}"
        )
        if runner.compression_rate_target is None:
            runner.compression_rate_step = np.abs(target_compression_rate - current_compression_rate)
            return target_compression_rate - current_compression_rate
        runner.compression_rate_step = np.abs(target_compression_rate - runner.compression_rate_target)
        return target_compression_rate - runner.compression_rate_target


class AccuracyAwareTrainingMode:
    EARLY_EXIT = "early_exit"
    ADAPTIVE_COMPRESSION_LEVEL = "adaptive_compression_level"


def create_accuracy_aware_training_loop(
    nncf_config: NNCFConfig,
    compression_ctrl: CompressionAlgorithmController,
    uncompressed_model_accuracy: float,
    **additional_runner_args: Any,
) -> BaseEarlyExitCompressionTrainingLoop:
    """
    Creates an accuracy aware training loop corresponding to NNCFConfig and CompressionAlgorithmController.
    :param: nncf_config: An instance of the NNCFConfig.
    :param: compression_ctrl: An instance of CompressionAlgorithmController.
    :param: uncompressed_model_accuracy: The value of the target accuracy metric for the original (uncompressed) model
    for purposes of matching the compressed model metric to this baseline
    :return: Accuracy aware training loop.
    """
    accuracy_aware_training_params = extract_accuracy_aware_training_params(nncf_config)
    accuracy_aware_training_mode = accuracy_aware_training_params.get("mode")
    if accuracy_aware_training_mode == AccuracyAwareTrainingMode.EARLY_EXIT:
        return EarlyExitCompressionTrainingLoop(
            nncf_config, compression_ctrl, uncompressed_model_accuracy, **additional_runner_args
        )
    if accuracy_aware_training_mode == AccuracyAwareTrainingMode.ADAPTIVE_COMPRESSION_LEVEL:
        return AdaptiveCompressionTrainingLoop(
            nncf_config,
            compression_ctrl,
            uncompressed_model_accuracy,
            **additional_runner_args,
        )
    raise nncf.InternalError("Incorrect accuracy aware mode in the config file")



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from pathlib import Path
from typing import Dict, List, Optional, Type

import jsonschema
import jstyleson as json

import nncf
from nncf.common.logging import nncf_logger
from nncf.common.utils.api_marker import api
from nncf.common.utils.os import safe_open
from nncf.config.definitions import SCHEMA_VISUALIZATION_URL
from nncf.config.schema import NNCF_CONFIG_SCHEMA
from nncf.config.schema import REF_VS_ALGO_SCHEMA
from nncf.config.schema import validate_single_compression_algo_schema
from nncf.config.structures import NNCFExtraConfigStruct


@api(canonical_alias="nncf.NNCFConfig")
class NNCFConfig(dict):
    """Contains the configuration parameters required for NNCF to apply the selected algorithms.

    This is a regular dictionary object extended with some utility functions, such as the ability to attach well-defined
    structures to pass non-serializable objects as parameters. It is primarily built from a .json file, or from a
    Python JSON-like dictionary - both data types will be checked against a JSONSchema. See the definition of the
    schema at https://openvinotoolkit.github.io/nncf/schema/, or by calling NNCFConfig.schema()."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__nncf_extra_structs: Dict[str, NNCFExtraConfigStruct] = {}

    @classmethod
    def from_dict(cls, nncf_dict: Dict) -> "NNCFConfig":
        """
        Load NNCF config from a Python dictionary. The dict must contain only JSON-supported primitives.

        :param nncf_dict: A Python dict with the JSON-style configuration for NNCF.
        """

        NNCFConfig.validate(nncf_dict)
        return cls(deepcopy(nncf_dict))

    @classmethod
    def from_json(cls, path: str) -> "NNCFConfig":
        """
        Load NNCF config from a JSON file at `path`.

        :param path: Path to the .json file containing the NNCF configuration.
        """
        file_path = Path(path)
        with safe_open(file_path) as f:
            loaded_json = json.load(f)
        return cls.from_dict(loaded_json)

    def register_extra_structs(self, struct_list: List[NNCFExtraConfigStruct]):
        """
        Attach the supplied list of extra configuration structures to this configuration object.

        :param struct_list: List of extra configuration structures.
        """
        for struct in struct_list:
            struct_id = struct.get_id()
            if struct_id in self.__nncf_extra_structs:
                raise nncf.InternalError(f"{struct_id} is already registered as extra struct in NNCFConfig!")
            self.__nncf_extra_structs[struct_id] = struct

    def get_extra_struct(self, struct_cls: Type[NNCFExtraConfigStruct]) -> NNCFExtraConfigStruct:
        return self.__nncf_extra_structs[struct_cls.get_id()]

    def has_extra_struct(self, struct_cls: Type[NNCFExtraConfigStruct]) -> NNCFExtraConfigStruct:
        return struct_cls.get_id() in self.__nncf_extra_structs

    def get_all_extra_structs(self) -> List[NNCFExtraConfigStruct]:
        return list(self.__nncf_extra_structs.values())

    def get_redefinable_global_param_value_for_algo(self, param_name: str, algo_name: str) -> Optional[str]:
        """
        Some parameters can be specified both on the global NNCF config .json level (so that they apply
        to all algos), and at the same time overridden in the algorithm-specific section of the .json.
        This function returns the value that should apply for a given algorithm name, considering the
        exact format of this config.

        :param param_name: The name of a parameter in the .json specification of the NNCFConfig, that may
          be present either at the top-most level of the .json, or at the top level of the algorithm-specific
          subdict.
        :param algo_name: The name of the algorithm (among the allowed algorithm names in the .json) for which
          the resolution of the redefinable parameter should occur.
        :return: The value of the parameter that should be applied for the algo specified by `algo_name`.
        """
        from nncf.config.extractors import extract_algo_specific_config

        algo_config = extract_algo_specific_config(self, algo_name)
        param = self.get(param_name)
        algo_specific_param = algo_config.get(param_name)
        if algo_specific_param is not None:
            param = algo_specific_param
        return param

    @staticmethod
    def schema() -> Dict:
        """
        Returns the JSONSchema against which the input data formats (.json or Python dict) are validated.
        """
        return NNCF_CONFIG_SCHEMA

    @staticmethod
    def _is_path_to_algorithm_name(path_parts: List[str]) -> bool:
        return (len(path_parts) == 2 and path_parts[0] == "compression" and path_parts[1] == "algorithm") or (
            len(path_parts) == 3
            and path_parts[0] == "compression"
            and path_parts[1].isnumeric()
            and path_parts[2] == "algorithm"
        )

    @staticmethod
    def validate(loaded_json):
        try:
            jsonschema.validate(loaded_json, NNCFConfig.schema())
        except jsonschema.ValidationError as e:
            nncf_logger.error("Invalid NNCF config supplied!")
            absolute_path_parts = [str(x) for x in e.absolute_path]
            if not NNCFConfig._is_path_to_algorithm_name(absolute_path_parts):
                e.message += f"\nRefer to the NNCF config schema documentation at {SCHEMA_VISUALIZATION_URL}"
                e.schema = "*schema too long for stdout display*"
                raise e

            # Need to make the error more algo-specific in case the config was so bad that no
            # scheme could be matched
            # If error is in the algo section, will revalidate the algo sections separately to
            # make the error message more targeted instead of displaying the entire huge schema.
            compression_section = loaded_json["compression"]
            if isinstance(compression_section, dict):
                validate_single_compression_algo_schema(compression_section, REF_VS_ALGO_SCHEMA)
            else:
                # Passed a list of dicts
                for compression_algo_dict in compression_section:
                    validate_single_compression_algo_schema(compression_algo_dict, REF_VS_ALGO_SCHEMA)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict

from nncf.common.graph.transformations.commands import TargetPoint
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.stateful_classes_registry import TF_STATEFUL_CLASSES


class TFTargetPointStateNames:
    OP_NAME = "op_name"
    OP_TYPE_NAME = "op_type_name"
    PORT_ID = "port_id"
    TARGET_TYPE = "target_type"


@TF_STATEFUL_CLASSES.register()
class TFTargetPoint(TargetPoint):
    """
    Describes where the compression operation should be placed.
    """

    _state_names = TFTargetPointStateNames

    def __init__(self, op_name: str, op_type_name: str, port_id: int, target_type: TargetType):
        """
        Initializes target point for TensorFlow backend.

        :param op_name: Name of a node in the `FuncGraph`.
        :param op_type_name: Type of operation.
        :param port_id: Port id.
        :param target_type: Type of the target point.
        """
        super().__init__(target_type)
        self.op_name = op_name
        self.op_type_name = op_type_name
        self.port_id = port_id

    def __eq__(self, other: "TFTargetPoint") -> bool:
        return (
            isinstance(other, TFTargetPoint)
            and self.type == other.type
            and self.op_name == other.op_name
            and self.op_type_name == other.op_type_name
            and self.port_id == other.port_id
        )

    def __str__(self) -> str:
        items = [
            super().__str__(),
            self.op_name,
            self.op_type_name,
            str(self.port_id),
        ]
        return " ".join(items)

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: State of the object.
        """
        state = {
            self._state_names.OP_NAME: self.op_name,
            self._state_names.OP_TYPE_NAME: self.op_type_name,
            self._state_names.PORT_ID: self.port_id,
            self._state_names.TARGET_TYPE: self.type.get_state(),
        }
        return state

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "TFTargetPoint":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        kwargs = {
            cls._state_names.OP_NAME: state[cls._state_names.OP_NAME],
            cls._state_names.OP_TYPE_NAME: state[cls._state_names.OP_TYPE_NAME],
            cls._state_names.PORT_ID: state[cls._state_names.PORT_ID],
            cls._state_names.TARGET_TYPE: TargetType.from_state(state[cls._state_names.TARGET_TYPE]),
        }
        return cls(**kwargs)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import csv
from abc import abstractmethod
from pathlib import Path
from typing import Any, Callable, Dict, NoReturn, Optional, Tuple, TypeVar

from nncf.common.logging import nncf_logger
from nncf.common.utils.os import safe_open

DataLoaderType = TypeVar("DataLoaderType")
TModel = TypeVar("TModel")
EvalFnType = Callable[[TModel], float]
AccValFnType = Callable[[TModel, DataLoaderType], float]


class BNASEvaluatorStateNames:
    BNAS_EVALUATOR_STAGE = "evaluator_state"


class BaseEvaluator:
    """
    An interface for handling measurements collected on a target device. Evaluators make use
    of functions provided by the users to measure a particular property, e.g., accuracy, latency, etc.
    """

    def __init__(self, name: str, ideal_val: float):
        """
        Initializes evaluator

        :param name: Name of the evaluator
        :param ideal_val: Ideal value for the metric computed by the evaluator
        """
        self.name = name
        self._current_value = -1
        self._ideal_value = ideal_val
        self.cache = {}
        # TODO(pablo): Here we should store some super-network signature that is associated with this evaluator

    @property
    def current_value(self):
        """
        :return: current value
        """
        return self._current_value

    @current_value.setter
    def current_value(self, val: float) -> NoReturn:
        """
        :param val: value to update the current value of the evaluator
        :return:
        """
        self._current_value = val

    def evaluate_and_add_to_cache_from_pymoo(self, pymoo_repr: Tuple[float, ...]) -> float:
        """
        Evaluates active sub-network and uses Pymoo representation for insertion in cache.

        :param model: Active sub-network
        :param pymoo_repr: tuple representing the associated values for the design variables
                            in Pymoo.
        :return: the value obtained from the model evaluation.
        """
        self._current_value = self.evaluate_subnet()
        self.add_to_cache(pymoo_repr, self._current_value)
        return self._current_value

    @abstractmethod
    def evaluate_subnet(self) -> float:
        """This method should implement how to a subnet is evaluated for a particular metric."""

    def add_to_cache(self, subnet_config_repr: Tuple[float, ...], measurement: float) -> NoReturn:
        """
        Adds evaluation result to cache

        :param subnet_config_repr: tuple containing the values for the associated design variables.
        :param measurement: value for the evaluator's metric.
        :return:
        """
        nncf_logger.debug(f"Add to evaluator {self.name}: {subnet_config_repr}, {measurement}")
        self.cache[subnet_config_repr] = measurement

    def retrieve_from_cache(self, subnet_config_repr: Tuple[float, ...]) -> Tuple[bool, float]:
        """
        Checks if sub-network info is in cache and returns the corresponding value.
        :param subnet_config_repr: tuple representing the values for the associated design variables.
        :return: (True if the information is in cache, and corresponding value stored in cache, 0 otherwise)
        """
        if subnet_config_repr in self.cache:
            return True, self.cache[subnet_config_repr]
        return False, 0

    def get_state(self) -> Dict[str, Any]:
        """
        Returns state of the evaluatar

        :return: Dict with the state of the evaluator
        """
        state_dict = {
            "name": self.name,
            "current_value": self._current_value,
            "ideal_value": self._ideal_value,
            "cache": self.cache,
        }
        return state_dict

    def update_from_state(self, state: Dict[str, Any]) -> NoReturn:
        """
        Updates the cache and other values in the evaluator from a saved state.

        :param state: dict with state that should be used for updating this evaluator
        :return:
        """
        new_dict = state.copy()
        self.name = new_dict["name"]
        self._ideal_value = new_dict["ideal_value"]
        self._current_value = new_dict["current_value"]
        self.cache = new_dict["cache"]

    def load_cache_from_csv(self, cache_file_path: str) -> NoReturn:
        """
        Loads cache from CSV file.

        :param cache_file_path: Path to CSV file containing the cache information.
        :return:
        """
        with safe_open(Path(cache_file_path), "r", encoding="utf8") as cache_file:
            reader = csv.reader(cache_file)
            for row in reader:
                rep_tuple = tuple(map(int, row[0][1 : len(row[0]) - 1].split(",")))
                self.add_to_cache(rep_tuple, float(row[1]))

    def export_cache_to_csv(self, cache_file_path: str) -> NoReturn:
        """
        Exports cache information to CSV.

        :param cache_file_path: Path to export a CSV file with the cache information.
        :return:
        """
        with safe_open(Path(cache_file_path) / f"cache_{self.name}.csv", "w", encoding="utf8") as cache_dump:
            writer = csv.writer(cache_dump)
            for key in self.cache:
                row = [key, self.cache[key]]
                writer.writerow(row)


class MACsEvaluator(BaseEvaluator):
    def __init__(self, eval_func):
        super().__init__("MACs", 0)
        self.eval_func = eval_func

    def evaluate_subnet(self) -> float:
        """
        Evaluates metric using model

        :param model: Active sub-network
        :return: value obtained from evaluation.
        """
        self._current_value = self.eval_func()
        return self._current_value


class AccuracyEvaluator(BaseEvaluator):
    """
    A particular kind of evaluator for collecting model's accuracy measurements
    """

    def __init__(
        self, model: TModel, eval_func: AccValFnType, val_loader: DataLoaderType, is_top1: Optional[bool] = True
    ):
        """
        Initializes Accuracy operator

        :param eval_func: function used to validate a sub-network
        :param val_loader: Datq loader used by the validation function
        :param is_top1: Whether is top 1 accuracy or top 5.
        :param ref_acc: Accuracy from a model that is used as input to BootstrapNAS
        """
        name = "top1_acc" if is_top1 else "top5_acc"
        super().__init__(name, -100)
        self._model = model
        self._eval_func = eval_func
        self._val_loader = val_loader
        self._is_top1 = is_top1

    def evaluate_subnet(self) -> float:
        """
        Obtain accuracy from evaluating the model.
        :param model: Active sub-network
        :return: accuracy from active sub-network.
        """
        self._current_value = self._eval_func(self._model, self._val_loader) * -1.0
        return self._current_value

    def get_state(self) -> Dict[str, Any]:
        """
        Get state of Accuracy evaluator.

        :return: Dict with state of evaluator
        """
        state = super().get_state()
        state["is_top1"] = self._is_top1
        return state

    def update_from_state(self, state: Dict[str, Any]) -> NoReturn:
        """

        :param state: dict with state that should be used for updating this evaluator
        :return:
        """

        super().update_from_state(state)
        new_dict = state.copy()
        self._is_top1 = new_dict["is_top1"]
